#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Misc."""
import io
from io import SEEK_CUR
from typing import Optional, Union

from spsdk import SPSDKError
from spsdk.utils.registers import value_to_int

from .header import Header


class RawDataException(SPSDKError):
    """Raw data read failed."""


class StreamReadFailed(RawDataException):
    """Read_raw_data could not read stream."""


class NotEnoughBytesException(RawDataException):
    """Read_raw_data could not read enough data."""


def hexdump_fmt(data: bytes, tab: int = 4, length: int = 16, sep: str = ":") -> str:
    """Dump some potentially larger data in hex."""
    text = " " * tab
    for i, j in enumerate(data):
        text += f"{j:02x}{sep}"
        if ((i + 1) % length) == 0:
            text += "\n" + " " * tab
    return text


def modulus_fmt(modulus: bytes, tab: int = 4, length: int = 15, sep: str = ":") -> str:
    """Modulus format."""
    return hexdump_fmt(b"\0" + modulus, tab, length, sep)


def read_raw_data(
    stream: Union[io.BufferedReader, io.BytesIO],
    length: int,
    index: Optional[int] = None,
    no_seek: bool = False,
) -> bytes:
    """Read raw data."""
    if index is not None:
        if index < 0:
            raise SPSDKError(f" Index must be non-negative, found {index}")
        if index != stream.tell():
            stream.seek(index)

    if length < 0:
        raise SPSDKError(f" Length must be non-negative, found {length}")

    try:
        data = stream.read(length)
    except Exception as exc:
        raise StreamReadFailed(f" stream.read() failed, requested {length} bytes") from exc

    if len(data) != length:
        raise NotEnoughBytesException(
            f" Could not read enough bytes, expected {length}, found {len(data)}"
        )

    if no_seek:
        stream.seek(-length, SEEK_CUR)

    return data


def read_raw_segment(
    buffer: Union[io.BufferedReader, io.BytesIO], segment_tag: int, index: Optional[int] = None
) -> bytes:
    """Read raw segment."""
    hrdata = read_raw_data(buffer, Header.SIZE, index)
    length = Header.parse(hrdata, 0, segment_tag).length - Header.SIZE
    return hrdata + read_raw_data(buffer, length)


def dict_diff(main: dict, mod: dict) -> dict:
    """Return a difference between two dictionaries if key is not present in main, it's skipped."""
    diff = {}
    for key, value in mod.items():
        if isinstance(value, dict):
            sub = dict_diff(main[key], value)
            if sub:
                diff[key] = sub
        else:
            if key not in main:
                continue
            main_value = main[key] if isinstance(main, dict) else main
            try:
                if value_to_int(main_value) != value_to_int(value):
                    diff[key] = value
            except (SPSDKError, TypeError):
                # Not a number!
                if main_value != value:
                    diff[key] = value
    return diff



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Trust Provisioning utilities."""
import os
from typing import List, Optional, Type

from spsdk.crypto import ec
from spsdk.exceptions import SPSDKError
from spsdk.utils.misc import load_configuration

from . import TP_DATA_FOLDER, TpDevInterface, TpIntfDescription, TpTargetInterface
from .adapters import TP_DEVICES, TP_TARGETS


def single_tp_device_adapter() -> bool:
    """Return True if there's only one TP device adapter."""
    return len(TP_DEVICES) == 1


def single_tp_target_adapter() -> bool:
    """Return True if there's only one TP target adapter."""
    return len(TP_TARGETS) == 1


def get_supported_devices() -> List[str]:
    """Return list of supported devices for Trust Provisioning.

    :return: List of devices.
    """
    data = load_configuration(os.path.join(TP_DATA_FOLDER, "database.yaml"))
    return list(data["devices"].keys())


def get_tp_device_types() -> List[str]:
    """Return list of supported TP device types."""
    return list(TP_DEVICES.keys())


def scan_tp_devices(
    tpdev: Optional[str] = None, settings: Optional[dict] = None
) -> List[TpIntfDescription]:
    """The function scans the TP devices on system.

    :param tpdev: Selection of one type of TP device, defaults to None (scan all supported).
    :param settings: Additional settings to setup interface, defaults to {}.
    :return: List of active TP device descriptors.
    :raises SPSDKError: Invalid value of parameter.
    """
    if tpdev and tpdev not in get_tp_device_types():
        raise SPSDKError(f"Unsupported TP device name - {tpdev}")

    dev_list = [tpdev] if tpdev else get_tp_device_types()

    active_devices = []
    for dev in dev_list:
        active_devices.extend(TP_DEVICES[dev].get_connected_devices(settings))

    return active_devices


def get_tp_devices(
    tpdev: Optional[str] = None, settings: Optional[dict] = None
) -> List[TpDevInterface]:
    """Return a list of active TP Devices fulfilling criteria in 'settings'.

    This functions attempts to open/close the device, please mind possible side-effects.

    :param tpdev: Name of TP Device interface, defaults to None
    :param settings: Settings for TP Target, defaults to None
    :return: List of active TP Devices
    """
    device_descriptors = scan_tp_devices(tpdev=tpdev, settings=settings)
    devices = []
    for descriptor in device_descriptors:
        try:
            dev = descriptor.create_interface()
            assert isinstance(dev, TpDevInterface)
            dev.open()
            dev.close()
            devices.append(dev)
        except Exception:  # pylint: disable=broad-except   # the underlying error is unknown
            dev.close()
            continue
    return devices


def get_tp_target_types() -> List[str]:
    """Return list of supported TP targets."""
    return list(TP_TARGETS.keys())


def scan_tp_targets(
    tptarget: Optional[str] = None, settings: Optional[dict] = None
) -> List[TpIntfDescription]:
    """The function scans the TP targets on system.

    :param tptarget: Selection of one type of TP target, defaults to None (scan all supported).
    :param settings: Additional settings to setup interface, defaults to {}.
    :return: List of active TP devices.
    :raises SPSDKError: Invalid value of parameter.
    """
    if tptarget and tptarget not in get_tp_target_types():
        raise SPSDKError(f"Unsupported TP device name - {tptarget}")

    target_list = [tptarget] if tptarget else get_tp_target_types()

    active_targets = []
    for target in target_list:
        active_targets.extend(TP_TARGETS[target].get_connected_targets(settings))

    return active_targets


def get_tp_targets(
    tptarget: Optional[str] = None, settings: Optional[dict] = None
) -> List[TpTargetInterface]:
    """Return a list of active TP Targets fulfilling criteria in 'settings'.

    This functions attempts to open/close the device, please mind possible side-effects.

    :param tptarget: Name of TP Target interface, defaults to None
    :param settings: Settings for TP Target, defaults to None
    :return: List is active TP Targets
    """
    target_descriptors = scan_tp_devices(tpdev=tptarget, settings=settings)
    targets = []
    for descriptor in target_descriptors:
        try:
            target = descriptor.create_interface()
            assert isinstance(target, TpTargetInterface)
            target.open()
            target.close()
            targets.append(target)
        except Exception:  # pylint: disable=broad-except   # the underlying error is unknown
            target.close()
            continue
    return targets


def get_tp_device_class(name: str) -> Type[TpDevInterface]:
    """Return class of TP device interface by name.

    :param name: Name of the interface.
    :return: TP device interface.
    """
    return TP_DEVICES[name]


def get_tp_target_class(name: str) -> Type[TpTargetInterface]:
    """Return class of TP target interface by name.

    :param name: Name of the interface.
    :return: TP target interface.
    """
    return TP_TARGETS[name]


def reconstruct_cryptography_key(key_material: bytes) -> ec.EllipticCurvePublicKey:
    """Reconstruct cryptography's ECC Public Key from coordinates."""
    coordinate_length = len(key_material) // 2
    curve = {32: ec.SECP256R1(), 48: ec.SECP384R1()}[coordinate_length]
    point_x = int.from_bytes(key_material[:coordinate_length], byteorder="big")
    point_y = int.from_bytes(key_material[coordinate_length:], byteorder="big")
    pub_numbers = ec.EllipticCurvePublicNumbers(x=point_x, y=point_y, curve=curve)
    key = pub_numbers.public_key()
    return key



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""DK6 Drivers backend interface."""
import logging
import time
from enum import Enum
from typing import List, Union

from spsdk import SPSDKError
from spsdk.dk6.serial_device import SerialDevice

logger = logging.getLogger(__name__)

DELAY_ISP_IO_TOGGLE_WAIT_MS = 100

# FTDI sequence in format PIN, BITMODE, MS_WAIT
FTDI_ISP_SEQUENCE = [
    (0xC0, 0x20, 10),
    (0xC4, 0x20, DELAY_ISP_IO_TOGGLE_WAIT_MS),
    (0xCC, 0x20, 10),
    (0x0C, 0x20, 10),
    (0x00, 0x00, 10),
]


def generate_pyftdi_url(serial_number: str, address: int = 1) -> str:
    """Generates URL for PyFTDI from serial number.

    :param serial_number: serial number from USBDeviceDescriptor
    :param address: interface number
    :return: url for PyFTDI
    """
    return f"ftdi://::{serial_number}/{address}"


class Backend(Enum):
    """Backend selection."""

    PYFTDI = 1
    PYLIBFTDI = 2
    FTD2xx = 3
    PYSERIAL = 4


class DeviceInfo:
    """Device info class.

    Contains information about the connected device
    """

    def __init__(
        self,
        device_id: Union[str, int],
        vid: int,
        pid: int,
        sn: str,
        description: str,
        address: int,
        backend: Backend,
    ):
        """Device info constructor.

        :param device_id: Device ID
        :param vid: USB VID
        :param pid: USB PID
        :param sn: Serial number
        :param description: description
        :param address: device address
        :param backend: backend
        """
        self.device_id = device_id
        self.vid = vid
        self.pid = pid
        self.sn = sn
        self.description = description
        self.address = address
        self.backend = backend

    def __str__(self) -> str:
        return (
            f"DEVICE ID: {self.device_id}, VID: {hex(self.vid)}, PID: {hex(self.pid)}, Serial number: {self.sn}, "
            + f"Description: {self.description}, Address: {self.address}, Backend: {self.backend}"
        )


class DriverInterface:
    """Interface to FTDI backends.

    Supported backends are: pyftdi, pylibftdi and ftdi2xx
    """

    def __init__(self, backend: Backend) -> None:
        """Initialize driver interface and serial interface based on backend.

        :param backend: supported backend name
        """
        self.backend = backend
        self.initialized = False

        logger.info(f"Initializing backend {backend}")
        if self.backend == Backend.PYFTDI:
            from pyftdi import ftdi

            self.driver = ftdi.Ftdi()
            self.dev = None

        elif self.backend == Backend.PYLIBFTDI:
            from pylibftdi import Device, Driver

            self.dev = Device()
            self.driver = Driver()
        else:
            self.dev = None
            self.driver = None

    def go_to_isp(self, device_id: str) -> None:
        """Send a sequence that goes to ISP mode using FTDI bitbang device."""
        logger.info("Sending bitbang sequence to ISP mode")
        if self.backend == Backend.PYFTDI:
            from pyftdi import ftdi

            url = generate_pyftdi_url(device_id)

            self.driver.open_bitbang_from_url(url)
            for ins in FTDI_ISP_SEQUENCE:
                bitmode = ftdi.Ftdi.BitMode.CBUS if ins[1] == 0x20 else ftdi.Ftdi.BitMode.RESET
                self.driver.set_bitmode(ins[0], bitmode)
                time.sleep(ins[2] / 1000)

        elif self.backend == Backend.FTD2xx:
            import ftd2xx as ftd

            if self.dev is None:
                logger.info("Initializing serial first before ISP bitbang for D2XX backend")

                self.dev = ftd.open(int(device_id))
            for ins in FTDI_ISP_SEQUENCE:
                self.dev.setBitMode(ins[0], ins[1])
                time.sleep(ins[2] / 1000)
            self.dev.close()

        else:
            logger.error("Selected backend does not have method for ISP sequence")

    def list_devices(self) -> List[DeviceInfo]:
        """Returns a list of devices that are connected for selected backend.

        :return: List of devices
        """
        devices_info = []
        logger.info("Enumerating DK6 devices")
        if self.backend == Backend.PYFTDI:
            devices_list = self.driver.list_devices()
            for device in devices_list:
                device_info = DeviceInfo(
                    device[0].sn,
                    device[0].vid,
                    device[0].pid,
                    device[0].sn,
                    device[0].description,
                    device[0].address,
                    self.backend,
                )
                devices_info.append(device_info)

        elif self.backend == Backend.FTD2xx:
            import ftd2xx as ftd

            device_count = ftd.createDeviceInfoList()
            for n in range(device_count):
                dev_dict = ftd.getDeviceInfoDetail(n)
                dev_info = DeviceInfo(
                    device_id=dev_dict.get("index"),
                    vid=0,
                    pid=0,
                    sn=dev_dict.get("serial"),
                    description=dev_dict.get("description"),
                    address=dev_dict.get("id"),
                    backend=self.backend,
                )
                devices_info.append(dev_info)

        elif self.backend == Backend.PYSERIAL:
            import serial.tools.list_ports

            ports = serial.tools.list_ports.comports()

            for port in ports:
                dev_info = DeviceInfo(
                    device_id=port.device,
                    vid=port.vid,
                    pid=port.pid,
                    sn=port.serial_number,
                    description=port.description,
                    address=port.location,
                    backend=self.backend,
                )
                devices_info.append(dev_info)

        else:
            logger.error("Selected backend does not implement method for listing devices")

        return devices_info

    def init_serial(self, device_id: str, baudrate: int = 115200, timeout: int = 5000) -> None:
        """Initialize serial device.

        :param device_id: device ID
        :param baudrate: UART baudrate, defaults to 115200
        :param timeout: read and write timeout, defaults to 5000 ms
        :raises SPSDKError: if invalid device_id is provided
        """
        logger.info(
            f"Initializing serial device for dev: {device_id}, baudrate: {baudrate} and timeout: {timeout}"
        )
        if self.initialized:
            logger.info("Serial already initialized, skipping initialization")
            return

        if self.backend == Backend.PYFTDI:
            import pyftdi.serialext

            timeout //= 1000
            url = generate_pyftdi_url(device_id)
            self.dev = pyftdi.serialext.serial_for_url(
                url, baudrate=baudrate, timeout=timeout, write_timeout=timeout
            )
            self.dev.reset_input_buffer()
            self.dev.reset_output_buffer()
            self.initialized = True

        elif self.backend == Backend.PYLIBFTDI:
            from pylibftdi import Device

            self.dev = Device(device_id=device_id)
            self.initialized = True

        elif self.backend == Backend.FTD2xx:
            import ftd2xx as ftd

            try:
                device_id_int = int(device_id)
            except ValueError as exc:
                raise SPSDKError("Invalid value for FTD2xx DEVICE_ID, it must be int") from exc
            self.dev = ftd.open(device_id_int)
            self.dev.resetPort()
            self.dev.setBaudRate(baudrate)
            self.dev.setTimeouts(timeout, timeout)
            self.initialized = True

        elif self.backend == Backend.PYSERIAL:
            import serial

            timeout //= 1000
            self.dev = serial.Serial(
                port=device_id,
                baudrate=baudrate,
                timeout=timeout,
                write_timeout=timeout,
            )
            self.initialized = True

    def get_serial(self) -> SerialDevice:
        """Return serial device.

        :raises SPSDKError: if serial device is not initialized
        :return: Serial device based on backend
        """
        if not self.initialized or not self.dev:
            raise SPSDKError("Serial device is not initialized")
        return self.dev



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for serial communication with a target device using MBoot protocol."""

import logging
from struct import pack, unpack_from
from typing import List, Optional, Union

import libusbsio

from spsdk.utils.exceptions import SPSDKTimeoutError
from spsdk.utils.misc import get_hash
from spsdk.utils.usbfilter import NXPUSBDeviceFilter, USBDeviceFilter

from ..commands import CmdPacket, CmdResponse, parse_cmd_response
from ..exceptions import McuBootConnectionError, McuBootDataAbortError, McuBootError
from .base import MBootInterface

logger = logging.getLogger(__name__)

# import os
# os.environ['PYUSB_DEBUG'] = 'debug'
# os.environ['PYUSB_LOG_FILENAME'] = 'usb.log'

REPORT_ID = {
    # USB HID Reports
    "CMD_OUT": 0x01,
    "CMD_IN": 0x03,
    "DATA_OUT": 0x02,
    "DATA_IN": 0x04,
}

########################################################################################################################
# Devices
########################################################################################################################

USB_DEVICES = {
    # NAME   | VID   | PID
    "MKL27": (0x15A2, 0x0073),
    "LPC55": (0x1FC9, 0x0021),
    "IMXRT": (0x1FC9, 0x0135),
    "MXRT20": (0x15A2, 0x0073),  # this is ID of flash-loader for RT102x
    "MXRT50": (0x15A2, 0x0073),  # this is ID of flash-loader for RT105x
    "MXRT60": (0x15A2, 0x0073),  # this is ID of flash-loader for RT106x
    "LPC55xx": (0x1FC9, 0x0020),
    "LPC551x": (0x1FC9, 0x0022),
    "RT6xx": (0x1FC9, 0x0021),
    "RT5xx_A": (0x1FC9, 0x0020),
    "RT5xx_B": (0x1FC9, 0x0023),
    "RT5xx_C": (0x1FC9, 0x0023),
    "RT5xx": (0x1FC9, 0x0023),
    "RT6xxM": (0x1FC9, 0x0024),
    "LPC553x": (0x1FC9, 0x0025),
    "MCXN9xx": (0x1FC9, 0x014F),
}


def scan_usb(device_id: Optional[str] = None) -> List["RawHid"]:
    """Scan connected USB devices.

    :param device_id: see USBDeviceFilter classes constructor for usb_id specification
    :return: list of matching RawHid devices
    """
    usb_filter = NXPUSBDeviceFilter(usb_id=device_id, nxp_device_names=USB_DEVICES)
    return RawHid.enumerate(usb_filter)


########################################################################################################################
# USB HID Interface Class
########################################################################################################################
class RawHid(MBootInterface):
    """Base class for OS specific RAW HID Interface classes."""

    @property
    def name(self) -> str:
        """Get the name of the device."""
        for name, value in USB_DEVICES.items():
            if value[0] == self.vid and value[1] == self.pid:
                return name
        return "Unknown"

    @property
    def is_opened(self) -> bool:
        """Return True if device is open."""
        return self.device is not None and self._opened

    @property
    def path_str(self) -> str:
        """BLHost-friendly string representation of USB path."""
        return NXPUSBDeviceFilter.convert_usb_path(self.path)

    @property
    def path_hash(self) -> str:
        """BLHost-friendly hash of the USB path."""
        return get_hash(self.path)

    def __init__(self) -> None:
        """Initialize the USB interface object."""
        super().__init__()
        self._opened = False
        self.vid = 0
        self.pid = 0
        self.serial_number = ""
        self.vendor_name = ""
        self.product_name = ""
        self.interface_number = 0
        self.timeout = 2000
        self.path = b""
        self.device: libusbsio.LIBUSBSIO.HID_DEVICE = None

    @staticmethod
    def _encode_report(report_id: int, data: bytes) -> bytes:
        """Encode the USB packet.

        :param report_id: ID of the report (see: HID_REPORT)
        :param data: Data to send
        :return: Encoded bytes and length of the final report frame
        """
        raw_data = pack("<2BH", report_id, 0x00, len(data))
        raw_data += data
        logger.debug(f"OUT[{len(raw_data)}]: {', '.join(f'{b:02X}' for b in raw_data)}")
        return raw_data

    @staticmethod
    def _decode_report(raw_data: bytes) -> Union[CmdResponse, bytes]:
        """Decodes the data read on USB interface.

        :param raw_data: Data received
        :return: CmdResponse object or data read
        :raises McuBootDataAbortError: Transaction aborted by target
        """
        logger.debug(f"IN [{len(raw_data)}]: {', '.join(f'{b:02X}' for b in raw_data)}")
        report_id, _, plen = unpack_from("<2BH", raw_data)
        if plen == 0:
            raise McuBootDataAbortError()
        data = raw_data[4 : 4 + plen]
        if report_id == REPORT_ID["CMD_IN"]:
            return parse_cmd_response(data)
        return data

    def info(self) -> str:
        """Return information about the USB interface."""
        return f"{self.product_name:s} (0x{self.vid:04X}, 0x{self.pid:04X}) path={self.path!r}"

    def open(self) -> None:
        """Open the interface.

        :raises McuBootConnectionError: if no device is available
        :raises McuBootConnectionError: if the device can not be opened
        """
        logger.debug("Open Interface")
        try:
            if not self.device:
                raise McuBootConnectionError("No device available")
            self.device.Open(self.path)
            self._opened = True
        except Exception as error:
            raise McuBootConnectionError(
                f"Unable to open device '{self.path!r}' VID={self.vid} PID={self.pid} SN='{self.serial_number}'"
            ) from error

    def close(self) -> None:
        """Close the interface.

        :raises McuBootConnectionError: if no device is available
        :raises McuBootConnectionError: if the device can not be opened
        """
        logger.debug("Close Interface")
        try:
            if not self.device:
                raise McuBootConnectionError("No device available")
            self.device.Close()
            self._opened = False
        except Exception as error:
            raise McuBootConnectionError(
                f"Unable to close device '{self.path!r}' VID={self.vid} PID={self.pid} SN='{self.serial_number}'"
            ) from error

    def write(self, packet: Union[CmdPacket, bytes]) -> None:
        """Write data on the OUT endpoint associated to the HID interfaces.

        :param packet: Data to send
        :raises McuBootError: Raises an error if packet type is incorrect
        :raises McuBootConnectionError: Raises an error if device is not opened for writing
        :raises McuBootConnectionError: Raises an error if device is not available
        :raises McuBootConnectionError: Raises an error if write operation fails
        :raises McuBootDataAbortError: May happen when ROM rejects SB file in shortened evaluation
        """
        if not self.device:
            raise McuBootConnectionError("No device available")
        if not self.is_opened:
            raise McuBootConnectionError("Device is opened for writing")

        if isinstance(packet, CmdPacket):
            report_id = REPORT_ID["CMD_OUT"]
            data = packet.to_bytes(padding=False)
        elif isinstance(packet, (bytes, bytearray)):
            report_id = REPORT_ID["DATA_OUT"]
            data = packet
        else:
            raise McuBootError("Packet has to be either 'CmdPacket' or 'bytes'")

        # try to read a begging of the ABORT_FRAME
        if self.allow_abort and report_id == REPORT_ID["DATA_OUT"]:
            try:
                (abort_data, _result) = self.device.Read(1024, timeout_ms=10)
                logger.debug(f"Read {len(abort_data)} bytes of abort data")
            except Exception as e:
                raise McuBootConnectionError(str(e)) from e
            if abort_data:
                logger.debug(f"{', '.join(f'{b:02X}' for b in abort_data)}")
                raise McuBootDataAbortError()

        try:
            raw_data = self._encode_report(report_id, data)
            _bytes_written = self.device.Write(raw_data, timeout_ms=self.timeout)

        except Exception as e:
            raise McuBootConnectionError(str(e)) from e

        if _bytes_written < 0 or _bytes_written < len(raw_data):
            raise McuBootConnectionError(
                f"Invalid size of written bytes has been detected: {_bytes_written} != {len(raw_data)}"
            )

    def read(self) -> Union[CmdResponse, bytes]:
        """Read data on the IN endpoint associated to the HID interface.

        :return: Return CmdResponse object.
        :raises McuBootConnectionError: Raises an error if device is not opened for reading
        :raises McuBootConnectionError: Raises if device is not available
        :raises McuBootConnectionError: Raises if reading fails
        :raises SPSDKTimeoutError: Time-out
        """
        if not self.is_opened:
            raise McuBootConnectionError("Device is not opened for reading")
        if not self.device:
            raise McuBootConnectionError("Device not available")
        try:
            (raw_data, result) = self.device.Read(1024, timeout_ms=self.timeout)
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e
        if not raw_data:
            logger.error(f"Cannot read from HID device, error={result}")
            raise SPSDKTimeoutError()
        # NOTE: uncomment the following when using KBoot/Flashloader v2.1 and older
        # import platform
        # if platform.system() == "Linux":
        #     raw_data += self.device.Read(1024, timeout_ms=self.timeout)
        return self._decode_report(bytes(raw_data))

    @staticmethod
    def enumerate(usb_device_filter: USBDeviceFilter) -> List["RawHid"]:
        """Get list of all connected devices matching the USBDeviceFilter object.

        :param usb_device_filter: USBDeviceFilter object
        :return: List of interfaces found
        """
        devices = []

        # use HID_API of LIBUSBSIO library to enumerate all USB HID devices
        libusbsio_logger = logging.getLogger("libusbsio")
        sio = libusbsio.usbsio(loglevel=libusbsio_logger.getEffectiveLevel())
        all_hid_devices = sio.HIDAPI_Enumerate()

        # iterate on all devices found
        for dev in all_hid_devices:
            if usb_device_filter.compare(vars(dev)) is True:
                new_device = RawHid()
                new_device.device = sio.HIDAPI_DeviceCreate()
                new_device.vid = dev["vendor_id"]
                new_device.pid = dev["product_id"]
                new_device.vendor_name = dev["manufacturer_string"]
                new_device.product_name = dev["product_string"]
                new_device.interface_number = dev["interface_number"]
                new_device.path = dev["path"]
                devices.append(new_device)

        return devices

    def __hash__(self) -> int:
        return hash(self.path)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Miscellaneous functions used throughout the SPSDK."""
import contextlib
import hashlib
import logging
import math
import os
import re
import time
from math import ceil
from struct import pack, unpack
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    Iterator,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
)

from spsdk import SPSDKError
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.exceptions import SPSDKTimeoutError

# for generics
T = TypeVar("T")  # pylint: disable=invalid-name

logger = logging.getLogger(__name__)


class BinaryPattern:
    """Binary pattern class.

    Supported patterns:
        - rand: Random Pattern
        - zeros: Filled with zeros
        - ones: Filled with all ones
        - inc: Filled with repeated numbers incremented by one 0-0xff
        - any kind of number, that will be repeated to fill up whole image.
          The format could be decimal, hexadecimal, bytes.
    """

    SPECIAL_PATTERNS = ["rand", "zeros", "ones", "inc"]

    def __init__(self, pattern: str) -> None:
        """Constructor of pattern class.

        :param pattern: Supported patterns:
                        - rand: Random Pattern
                        - zeros: Filled with zeros
                        - ones: Filled with all ones
                        - inc: Filled with repeated numbers incremented by one 0-0xff
                        - any kind of number, that will be repeated to fill up whole image.
                        The format could be decimal, hexadecimal, bytes.
        :raises SPSDKValueError: Unsupported pattern detected.
        """
        try:
            value_to_int(pattern)
        except SPSDKError:
            if not pattern in BinaryPattern.SPECIAL_PATTERNS:
                raise SPSDKValueError(  # pylint: disable=raise-missing-from
                    f"Unsupported input pattern {pattern}"
                )

        self._pattern = pattern

    def get_block(self, size: int) -> bytes:
        """Get block filled with pattern.

        :param size: Size of block to return.
        :return: Filled up block with specified pattern.
        """
        if self._pattern == "zeros":
            return bytes(size)

        if self._pattern == "ones":
            return bytes(b"\xff" * size)

        if self._pattern == "rand":
            # pylint: disable=import-outside-toplevel
            from spsdk.utils.crypto.common import crypto_backend

            return crypto_backend().random_bytes(size)

        if self._pattern == "inc":
            return bytes((x & 0xFF for x in range(size)))

        pattern = value_to_bytes(self._pattern, align_to_2n=False)
        block = bytes(pattern * (int((size / len(pattern))) + 1))
        return block[:size]

    @property
    def pattern(self) -> str:
        """Get the pattern.

        :return: Pattern in string representation.
        """
        try:
            return hex(value_to_int(self._pattern))
        except SPSDKError:
            return self._pattern


def align(number: int, alignment: int = 4) -> int:
    """Align number (size or address) size to specified alignment, typically 4, 8 or 16 bytes boundary.

    :param number: input to be aligned
    :param alignment: the boundary to align; typical value is power of 2
    :return: aligned number; result is always >= size (e.g. aligned up)
    :raises SPSDKError: When there is wrong alignment
    """
    if alignment <= 0 or number < 0:
        raise SPSDKError("Wrong alignment")

    return (number + (alignment - 1)) // alignment * alignment


def align_block(
    data: Union[bytes, bytearray],
    alignment: int = 4,
    padding: Optional[Union[int, str, BinaryPattern]] = None,
) -> bytes:
    """Align binary data block length to specified boundary by adding padding bytes to the end.

    :param data: to be aligned
    :param alignment: boundary alignment (typically 2, 4, 16, 64 or 256 boundary)
    :param padding: byte to be added or BinaryPattern
    :return: aligned block
    :raises SPSDKError: When there is wrong alignment
    """
    assert isinstance(data, (bytes, bytearray))

    if alignment < 0:
        raise SPSDKError("Wrong alignment")
    current_size = len(data)
    num_padding = align(current_size, alignment) - current_size
    if not num_padding:
        return bytes(data)
    if not padding:
        padding = BinaryPattern("zeros")
    elif not isinstance(padding, BinaryPattern):
        padding = BinaryPattern(str(padding))
    return bytes(data + padding.get_block(num_padding))


def align_block_fill_random(data: bytes, alignment: int = 4) -> bytes:
    """Same as `align_block`, just parameter `padding` is fixed to `-1` to fill with random data."""
    return align_block(data, alignment, BinaryPattern("rand"))


def extend_block(data: bytes, length: int, padding: int = 0) -> bytes:
    """Add padding to the binary data block to extend the length to specified value.

    :param data: block to be extended
    :param length: requested block length; the value must be >= current block length
    :param padding: 8-bit value value to be used as a padding
    :return: block extended with padding
    :raises SPSDKError: When the length is incorrect
    """
    current_len = len(data)
    if length < current_len:
        raise SPSDKError("Incorrect length")
    num_padding = length - current_len
    if not num_padding:
        return data
    return data + bytes([padding]) * num_padding


def find_first(iterable: Iterable[T], predicate: Callable[[T], bool]) -> Optional[T]:
    """Find first element from the list, that matches the condition.

    :param iterable: list of elements
    :param predicate: function for selection of the element
    :return: found element; None if not found
    """
    return next((a for a in iterable if predicate(a)), None)


def load_binary(path: str, search_paths: Optional[List[str]] = None) -> bytes:
    """Loads binary file into bytes.

    :param path: Path to the file.
    :param search_paths: List of paths where to search for the file, defaults to None
    :return: content of the binary file as bytes
    """
    data = load_file(path, mode="rb", search_paths=search_paths)
    assert isinstance(data, bytes)
    return data


def load_text(path: str, search_paths: Optional[List[str]] = None) -> str:
    """Loads binary file into bytes.

    :param path: Path to the file.
    :param search_paths: List of paths where to search for the file, defaults to None
    :return: content of the binary file as bytes
    """
    text = load_file(path, mode="r", search_paths=search_paths)
    assert isinstance(text, str)
    return text


def load_file(
    path: str, mode: str = "r", search_paths: Optional[List[str]] = None
) -> Union[str, bytes]:
    """Loads a file into bytes.

    :param path: Path to the file.
    :param mode: mode for reading the file 'r'/'rb'
    :param search_paths: List of paths where to search for the file, defaults to None
    :return: content of the binary file as bytes or str (based on mode)
    """
    path = find_file(path, search_paths=search_paths)
    logger.debug(f"Loading {'binary' if 'b' in mode else 'text'} file from {path}")
    with open(path, mode) as f:
        return f.read()


def write_file(
    data: Union[str, bytes], path: str, mode: str = "w", encoding: Optional[str] = None
) -> int:
    """Writes data into a file.

    :param data: data to write
    :param path: Path to the file.
    :param mode: writing mode, 'w' for text, 'wb' for binary data, defaults to 'w'
    :param encoding: Encoding of written file ('ascii', 'utf-8').
    :return: number of written elements
    """
    path = path.replace("\\", "/")
    folder = os.path.dirname(path)
    if folder and not os.path.exists(folder):
        os.makedirs(folder, exist_ok=True)

    logger.debug(f"Storing {'binary' if 'b' in mode else 'text'} file at {path}")
    with open(path, mode, encoding=encoding) as f:
        return f.write(data)


def get_abs_path(file_path: str, base_dir: Optional[str] = None) -> str:
    """Return a full path to the file.

    param base_dir: Base directory to create absolute path, if not specified the system CWD is used.
    return: Absolute file path.
    """
    if os.path.isabs(file_path):
        return file_path.replace("\\", "/")

    return os.path.abspath(os.path.join(base_dir or os.getcwd(), file_path)).replace("\\", "/")


def find_file(
    file_path: str,
    use_cwd: bool = True,
    search_paths: Optional[List[str]] = None,
    raise_exc: bool = True,
) -> str:
    """Return a full path to the file.

    `search_paths` takes precedence over `CWD` if used (default)

    :param file_path: File name, part of file path or full path
    :param use_cwd: Try current working directory to find the file, defaults to True
    :param search_paths: List of paths where to search for the file, defaults to None
    :param raise_exc: Raise exception if file is not found, defaults to True
    :return: Full path to the file
    :raises SPSDKError: File not found
    """
    file_path = file_path.replace("\\", "/")

    if os.path.isabs(file_path):
        return file_path
    if search_paths:
        for dir_candidate in search_paths:
            if not dir_candidate:
                continue
            dir_candidate = dir_candidate.replace("\\", "/")
            path_candidate = get_abs_path(file_path, base_dir=dir_candidate)
            if os.path.isfile(path_candidate):
                return path_candidate
    if use_cwd and os.path.isfile(file_path):
        return get_abs_path(file_path)
    # list all directories in error message
    searched_in: List[str] = []
    if use_cwd:
        searched_in.append(os.path.abspath(os.curdir))
    if search_paths:
        searched_in.extend(filter(None, search_paths))
    searched_in = [s.replace("\\", "/") for s in searched_in]
    err_str = f"File '{file_path}' not found, Searched in: {', '.join(searched_in)}"
    if not raise_exc:
        logger.debug(err_str)
        return ""  # TODO Maybe, it will be better return None and solve raised MyPY recommendations
    raise SPSDKError(err_str)


@contextlib.contextmanager
def use_working_directory(path: str) -> Iterator[None]:
    # pylint: disable=missing-yield-doc
    """Execute the block in given directory.

    Cd into specific directory.
    Execute the block.
    Change the directory back into the original one.

    :param path: the path, where the current directory will be changed to
    """
    current_dir = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(current_dir)
        assert os.getcwd() == current_dir


class DebugInfo:
    """The class is used to provide detailed information about export process and exported data.

    It is handy for analyzing content and debugging changes in the exported binary output.
    """

    @classmethod
    def disabled(cls) -> "DebugInfo":
        """Return an instance of DebugInfo with disabled message collecting."""
        return DebugInfo(enabled=False)

    def __init__(self, enabled: bool = True):
        """Constructor.

        :param enabled: True if logging enabled; False otherwise
        """
        self._lines: Optional[List[str]] = [] if enabled else None

    @property
    def enabled(self) -> bool:
        """:return: whether debugging enabled."""
        return self._lines is not None

    def append(self, line: str) -> None:
        """Appends the line to the log.

        :param line: text to be added
        :raises SPSDKError: When there is nothing to append
        """
        if self.enabled:
            if self._lines is None:
                raise SPSDKError("There is nothing to append")
            self._lines.append(line)

    def append_section(self, name: str) -> None:
        """Append new section to the debug log.

        :param name: of the section
        """
        self.append(f"[{name}]")

    def append_hex_data(self, data: bytes) -> None:
        """Append binary data in HEX form.

        :param data: to be logged
        """
        self.append("hex=" + data.hex())
        self.append("len=" + str(len(data)) + "=" + hex(len(data)))

    def append_binary_section(self, section_name: str, data: bytes) -> None:
        """Append section and binary data.

        :param section_name: the name
        :param data: binary data
        """
        self.append_section(section_name)
        self.append_hex_data(data)

    def append_binary_data(self, data_name: str, data: bytes) -> None:
        """Append short section with binary data.

        :param data_name: the name
        :param data: binary data (up to 8 bytes)
        :raises SPSDKError: When the data has incorrect length
        """
        if len(data) > 16:
            raise SPSDKError("Incorrect data length")
        self.append(data_name + "=" + data.hex())

    @property
    def lines(self) -> Iterable[str]:
        """:return: list of logged lines; empty list if nothing logged or log disabled."""
        if self._lines:
            return self._lines
        return []

    def info(self) -> str:
        """:return: multi-line text with log; empty string if nothing logged or log disabled."""
        return "\n".join(self.lines)


def format_value(value: int, size: int, delimiter: str = "_", use_prefix: bool = True) -> str:
    """Convert the 'value' into either BIN or HEX string, depending on 'size'.

    if 'size' is divisible by 8, function returns HEX, BIN otherwise
    digits in result string are grouped by 4 using 'delimiter' (underscore)
    """
    padding = size if size % 8 else (size // 8) * 2
    infix = "b" if size % 8 else "x"
    sign = "-" if value < 0 else ""
    parts = re.findall(".{1,4}", f"{abs(value):0{padding}{infix}}"[::-1])
    rev = delimiter.join(parts)[::-1]
    prefix = f"0{infix}" if use_prefix else ""
    return f"{sign}{prefix}{rev}"


def get_bytes_cnt_of_int(
    value: int, align_to_2n: bool = True, byte_cnt: Optional[int] = None
) -> int:
    """Returns count of bytes needed to store handled integer.

    :param value: Input integer value.
    :param align_to_2n: The result will be aligned to standard sizes 1,2,4,8,12,16,20.
    :param byte_cnt: The result count of bytes.
    :raises SPSDKValueError: The integer input value doesn't fit into byte_cnt.
    :return: Number of bytes needed to store integer.
    """
    cnt = 0
    if value == 0:
        return byte_cnt or 1

    while value != 0:
        value >>= 8
        cnt += 1

    if align_to_2n and cnt > 2:
        cnt = int(ceil(cnt / 4)) * 4

    if byte_cnt and cnt > byte_cnt:
        raise SPSDKValueError(
            f"Value takes more bytes than required byte count {byte_cnt} after align."
        )

    cnt = byte_cnt or cnt

    return cnt


def value_to_int(value: Union[bytes, bytearray, int, str], default: Optional[int] = None) -> int:
    """Function loads value from lot of formats to integer.

    :param value: Input value.
    :param default: Default Value in case of invalid input.
    :return: Value in Integer.
    :raises SPSDKError: Unsupported input type.
    """
    if isinstance(value, int):
        return value

    if isinstance(value, (bytes, bytearray)):
        return int.from_bytes(value, "big")

    if isinstance(value, str) and value != "":
        match = re.match(
            r"(?P<prefix>0[box])?(?P<number>[0-9a-f_]+)(?P<suffix>[ul]{0,3})$",
            value.strip().lower(),
        )
        if match:
            base = {"0b": 2, "0o": 8, "0": 10, "0x": 16, None: 10}[match.group("prefix")]
            try:
                return int(match.group("number"), base=base)
            except ValueError:
                pass

    if default is not None:
        return default
    raise SPSDKError(f"Invalid input number type({type(value)}) with value ({value})")


def value_to_bytes(
    value: Union[bytes, bytearray, int, str],
    align_to_2n: bool = True,
    byte_cnt: Optional[int] = None,
    endianness: str = "big",
) -> bytes:
    """Function loads value from lot of formats.

    :param value: Input value.
    :param align_to_2n: When is set, the function aligns length of return array to 1,2,4,8,12 etc.
    :param byte_cnt: The result count of bytes.
    :param endianness: The result bytes endianness ['big', 'little'].
    :return: Value in bytes.
    """
    if isinstance(value, bytes):
        return value

    if isinstance(value, bytearray):
        return bytes(value)

    value = value_to_int(value)
    return value.to_bytes(
        get_bytes_cnt_of_int(value, align_to_2n, byte_cnt=byte_cnt),
        endianness,  # type: ignore[arg-type]
    )


def value_to_bool(value: Union[bool, int, str]) -> bool:
    """Function decode bool value from various formats.

    :param value: Input value.
    :return: Boolean value.
    :raises SPSDKError: Unsupported input type.
    """
    if isinstance(value, bool):
        return value

    if isinstance(value, int):
        return bool(value)

    if isinstance(value, str):
        return value in ("True", "T", "1")

    raise SPSDKError(f"Invalid input Boolean type({type(value)}) with value ({value})")


def reverse_bytes_in_longs(arr: bytes) -> bytes:
    """The function reverse byte order in longs from input bytes.

    :param arr: Input array.
    :return: New array with reversed bytes.
    :raises SPSDKError: Raises when invalid value is in input.
    """
    arr_len = len(arr)
    if arr_len % 4 != 0:
        raise SPSDKError("The input array is not in modulo 4!")

    result = bytearray()

    for x in range(0, arr_len, 4):
        word = bytearray(arr[x : x + 4])
        word.reverse()
        result.extend(word)
    return bytes(result)


def reverse_bits_in_bytes(arr: bytes) -> bytes:
    """The function reverse bits order in input bytes.

    :param arr: Input array.
    :return: New array with reversed bits in bytes.
    :raises SPSDKError: Raises when invalid value is in input.
    """
    result = bytearray()

    for x in arr:
        result.append(int(f"{x:08b}"[::-1], 2))

    return bytes(result)


def change_endianness(bin_data: bytes) -> bytes:
    """Convert binary format used in files to binary used in register object.

    :param bin_data: input binary array.
    :return: Converted array (practically little to big endianness).
    :raises SPSDKError: Invalid value on input.
    """
    data = bytearray(bin_data)
    length = len(data)
    if length == 1:
        return data

    if length == 2:
        data.reverse()
        return data

    # The length of 24 bits is not supported yet
    if length == 3:
        raise SPSDKError("Unsupported length (3) for change endianness.")

    return reverse_bytes_in_longs(data)


class Timeout:
    """Simple timeout handle class."""

    UNITS = {
        "s": 1000000,
        "ms": 1000,
        "us": 1,
    }

    def __init__(self, timeout: int, units: str = "s") -> None:
        """Simple timeout class constructor.

        :param timeout: Timeout value.
        :param units: Timeout units (MUST be from the UNITS list)
        :raises SPSDKValueError: Invalid input value.
        """
        if units not in self.UNITS:
            raise SPSDKValueError("Units are not in supported units.")
        self.enabled = timeout != 0
        self.timeout_us = timeout * self.UNITS[units]
        self.start_time_us = self._get_current_time_us()
        self.end_time = self.start_time_us + self.timeout_us
        self.units = units

    @staticmethod
    def _get_current_time_us() -> int:
        """Returns current system time in microseconds.

        :return: Current time in microseconds
        """
        return ceil(time.time() * 1_000_000)

    def _convert_to_units(self, time_us: int) -> int:
        """Converts time in us into used units.

        :param time_us: Time in micro seconds.
        :return: Time in user units.
        """
        return time_us // self.UNITS[self.units]

    def get_consumed_time(self) -> int:
        """Returns consumed time since start of timeout operation.

        :return: Consumed time in units as the class was constructed
        """
        return self._convert_to_units(self._get_current_time_us() - self.start_time_us)

    def get_consumed_time_ms(self) -> int:
        """Returns consumed time since start of timeouted operation in milliseconds.

        :return: Consumed time in milliseconds
        """
        return (self._get_current_time_us() - self.start_time_us) // 1000

    def get_rest_time(self, raise_exc: bool = False) -> int:
        """Returns rest time to timeout overflow.

        :param raise_exc: If set, the function raise SPSDKTimeoutError in case of overflow.
        :return: Rest time in units as the class was constructed
        :raises SPSDKTimeoutError: In case of overflow
        """
        if self.enabled and self._get_current_time_us() > self.end_time and raise_exc:
            raise SPSDKTimeoutError("Timeout of operation.")

        return (
            self._convert_to_units(self.end_time - self._get_current_time_us())
            if self.enabled
            else 0
        )

    def get_rest_time_ms(self, raise_exc: bool = False) -> int:
        """Returns rest time to timeout overflow.

        :param raise_exc: If set, the function raise SPSDKTimeoutError in case of overflow.
        :return: Rest time in milliseconds
        :raises SPSDKTimeoutError: In case of overflow
        """
        if self.enabled and self._get_current_time_us() > self.end_time and raise_exc:
            raise SPSDKTimeoutError("Timeout of operation.")

        # pylint: disable=superfluous-parens     # because PEP20: Readability counts
        return ((self.end_time - self._get_current_time_us()) // 1000) if self.enabled else 0

    def overflow(self, raise_exc: bool = False) -> bool:
        """Check the the timer has been overflowed.

        :param raise_exc: If set, the function raise SPSDKTimeoutError in case of overflow.
        :return: True if timeout overflowed, False otherwise.
        :raises SPSDKTimeoutError: In case of overflow
        """
        overflow = self.enabled and self._get_current_time_us() > self.end_time
        if overflow and raise_exc:
            raise SPSDKTimeoutError("Timeout of operation.")
        return overflow


def size_fmt(num: Union[float, int], use_kibibyte: bool = True) -> str:
    """Size format."""
    base, suffix = [(1000.0, "B"), (1024.0, "iB")][use_kibibyte]
    i = "B"
    for i in ["B"] + [i + suffix for i in list("kMGTP")]:
        if num < base:
            break
        num /= base

    return f"{int(num)} {i}" if i == "B" else f"{num:3.1f} {i}"


def numberify_version(version: str, separator: str = ".", valid_numbers: int = 3) -> int:
    """Turn version string into a number.

    Each group is weighted by a multiple of 1000

    1.2.3    -> 1  * 1_000_000 +   2 * 1_000 + 3 * 1 =  1_002_003
    21.100.9 -> 21 * 1_000_000 + 100 * 1_000 + 9 * 1 = 21_100_009

    :param version: Version string numbers separated by `separator`
    :param separator: Separator used in the version string, defaults to "."
    :param valid_numbers: Amount of numbers to sanitize to consider, defaults to 3
    :return: Number representing the version
    """
    sanitized_version = sanitize_version(
        version=version, separator=separator, valid_numbers=valid_numbers
    )
    return int(
        sum(
            int(number) * math.pow(10, 3 * order)
            for order, number in enumerate(reversed(sanitized_version.split(separator)))
        )
    )


def sanitize_version(version: str, separator: str = ".", valid_numbers: int = 3) -> str:
    """Sanitize version string.

    Append '.0' in case version string has fewer parts than `valid_numbers`
    Remove right-most version parts after `valid_numbers` amount of parts

    1.2     -> 1.2.0
    1.2.3.4 -> 1.2.3

    :param version: Original version string
    :param separator: Separator used in the version string, defaults to "."
    :param valid_numbers: Amount of numbers to sanitize, defaults to 3
    :return: Sanitized version string
    """
    version_parts = version.split(separator)
    version_parts += ["0"] * (valid_numbers - len(version_parts))
    return separator.join(version_parts[:valid_numbers])


def get_key_by_val(value: str, dictionary: Dict[str, List[str]]) -> str:
    """Return key by its value.

    :param value: Value to find.
    :param dictionary: Dictionary to find in.
    :raises SPSDKValueError: Value is not present in dictionary.
    :return: Key name
    """
    for key, item in dictionary.items():
        if value.lower() in [x.lower() for x in item]:
            return key

    raise SPSDKValueError(f"Value {value} is not in {dictionary}.")


def swap16(x: int) -> int:
    """Swap bytes in half word (16bit).

    :param x: Original number
    :return: Number with swapped bytes
    :raises SPSDKError: When incorrect number to be swapped is provided
    """
    if x < 0 or x > 0xFFFF:
        raise SPSDKError("Incorrect number to be swapped")
    return ((x << 8) & 0xFF00) | ((x >> 8) & 0x00FF)


def swap32(x: int) -> int:
    """Swap 32 bit integer.

    :param x: integer to be swapped
    :return: swapped value
    :raises SPSDKError: When incorrect number to be swapped is provided
    """
    if x < 0 or x > 0xFFFFFFFF:
        raise SPSDKError("Incorrect number to be swapped")
    return unpack("<I", pack(">I", x))[0]


def check_range(x: int, start: int = 0, end: int = (1 << 32) - 1) -> bool:
    """Check if the number is in range.

    :param x: Number to check.
    :param start: Lower border of range, default is 0.
    :param end: Upper border of range, default is unsigned 32-bit range.
    :return: True if fits, False otherwise.
    """
    if start > x > end:
        return False

    return True


def load_configuration(path: str) -> dict:
    """Load configuration from yml/json file.

    :param path: Path to configuration file
    :raises SPSDKError: When unsupported file is provided
    :return: Content of configuration as dictionary
    """
    if not os.path.exists(path):
        raise SPSDKError(f"File not found'{path}'.")

    # import YAML only if needed to save startup time
    from ruamel.yaml import YAML, YAMLError  # pylint: disable=import-outside-toplevel

    try:
        with open(path) as f:
            return YAML(typ="safe").load(f)
    except (YAMLError, UnicodeDecodeError):
        pass

    # import json only if needed to save startup time
    import commentjson as json  # pylint: disable=import-outside-toplevel

    try:
        with open(path) as f:
            return json.load(f)
    except json.JSONLibraryException:
        pass

    raise SPSDKError(f"Unable to load '{path}'.")


def split_data(data: Union[bytearray, bytes], size: int) -> Generator[bytes, None, None]:
    """Split data into chunks of size.

    :param bytearray data: array of bytes to be split
    :param int size: size of splitted array
    :return Generator[bytes]: splitted array
    """
    for i in range(0, len(data), size):
        yield data[i : i + size]


def get_hash(text: Union[str, bytes]) -> str:
    """Returns hash of given text."""
    if isinstance(text, str):
        text = text.encode("utf-8")
    return hashlib.sha1(text).digest().hex()[:8]


TS = TypeVar("TS", bound="SingletonMeta")  # pylint: disable=invalid-name


class SingletonMeta(type):
    """Singleton metaclass."""

    _instance = None

    def __call__(cls: Type[TS], *args: Any, **kwargs: Any) -> TS:  # type: ignore
        """Call dunder override."""
        if cls._instance is None:
            instance = super().__call__(*args, **kwargs)
            cls._instance = instance
        return cls._instance



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for serial communication with a FRDM target device using BUSPAL protocol."""

import logging
import struct
import time
from enum import Enum
from typing import Any, Dict, List, Optional

from spsdk.exceptions import SPSDKError

from .base import Interface
from .buspal import BBConstants, Buspal, BuspalMode

logger = logging.getLogger(__name__)


def scan_buspal_spi(
    port: Optional[str] = None,
    timeout: int = Buspal.DEFAULT_TIMEOUT,
    props: Optional[List[str]] = None,
) -> List[Interface]:
    """Scan connected serial ports and set BUSPAL properties.

    Returns list of serial ports with devices that respond to BUSPAL communication protocol.
    If 'port' is specified, only that serial port is checked
    If no devices are found, return an empty list.

    :param port: name of preferred serial port, defaults to None
    :param timeout: timeout in milliseconds
    :param props: buspal target properties
    :return: list of available interfaces
    """
    return BuspalSPI.scan_buspal(port, timeout, props)


# pylint: disable=invalid-name
class SpiModeCommand(Enum):
    """Spi mode commands."""

    exit = 0x00  # 00000000 - Exit to bit bang mode
    version = 0x01  # 00000001 - Enter raw SPI mode, display version string
    chip_select = 0x02  # 0000001x - CS high (1) or low (0)
    sniff = 0x0C  # 000011XX - Sniff SPI traffic when CS low(10)/all(01)
    bulk_transfer = 0x10  # 0001xxxx - Bulk SPI transfer, send/read 1-16 bytes (0=1byte!)
    config_periph = 0x40  # 0100wxyz - Configure peripherals w=power, x=pull-ups, y=AUX, z=CS
    set_speed = 0x60  # 01100xxx - SPI speed
    config_spi = 0x80  # 1000wxyz - SPI config, w=HiZ/3.3v, x=CKP idle, y=CKE edge, z=SMP sample
    write_then_read = 0x04  # 00000100 - Write then read extended command


# pylint: disable=invalid-name
class SpiConfigShift(Enum):
    """Spi configuration shifts for the mask."""

    direction = 0
    phase = 1
    polarity = 2


# pylint: disable=invalid-name
class SpiClockPolarity(Enum):
    """SPI clock polarity configuration."""

    active_high = 0  # Active-high SPI clock (idles low).
    active_low = 1  # Active-low SPI clock (idles high).


# pylint: disable=invalid-name
class SpiClockPhase(Enum):
    """SPI clock phase configuration."""

    # First edge on SPSCK occurs at the middle of the first cycle of a data transfer.
    first_edge = 0
    # First edge on SPSCK occurs at the start of the first cycle of a data transfer.
    second_edge = 1


# pylint: disable=invalid-name
class SpiShiftDirection(Enum):
    """SPI clock phase configuration."""

    msb_first = 0  # Data transfers start with most significant bit.
    lsb_first = 1  # Data transfers start with least significant bit.


class SpiConfiguration:
    """Dataclass to store SPI configuration."""

    speed: int
    polarity: SpiClockPolarity
    phase: SpiClockPhase
    direction: SpiShiftDirection


class BuspalSPI(Buspal):
    """BUSPAL SPI interface."""

    TARGET_SETTINGS = ["speed", "polarity", "phase", "direction"]

    HDR_FRAME_RETRY_CNT = 3
    ACK_WAIT_DELAY = 0.01  # in seconds

    def __init__(self, port: str, timeout: int):
        """Initialize the BUSPAL SPI interface.

        :param port: name of the serial port, defaults to None
        :param timeout: read/write timeout in milliseconds
        """
        self.mode = BuspalMode.spi
        super().__init__(port, timeout)

    def configure(self, props: List[str]) -> None:
        """Configure the BUSPAL SPI interface.

        :param props: buspal settings
        """
        spi_props: Dict[str, Any] = dict(zip(self.TARGET_SETTINGS, props))

        speed = int(spi_props.get("speed", 100))
        polarity = SpiClockPolarity(spi_props.get("polarity", SpiClockPolarity.active_low))
        phase = SpiClockPhase(spi_props.get("phase", SpiClockPhase.second_edge))
        direction = SpiShiftDirection(spi_props.get("direction", SpiShiftDirection.msb_first))

        # set SPI config
        logger.debug("Set SPI config")
        spi_data = polarity.value << SpiConfigShift.polarity.value
        spi_data |= phase.value << SpiConfigShift.phase.value
        spi_data |= direction.value << SpiConfigShift.direction.value
        spi_data |= SpiModeCommand.config_spi.value
        self._send_command_check_response(bytes([spi_data]), bytes([BBConstants.response_ok.value]))

        # set SPI speed
        logger.debug(f"Set SPI speed to {speed}bps")
        spi_speed = struct.pack("<BI", SpiModeCommand.set_speed.value, speed)
        self._send_command_check_response(spi_speed, bytes([BBConstants.response_ok.value]))

    def _send_frame(
        self, frames: bytes, wait_for_ack: bool = True, retry_cnt: int = HDR_FRAME_RETRY_CNT
    ) -> None:
        """Send a frame to BUSPAL SPI device.

        :param frames: Data to send
        :param wait_for_ack: Wait for ACK frame from device, defaults to True
        :param retry_cnt: Number of retry in case the header frame is incorrect
        :raises AssertionError: Unexpected frame header or frame type (if specified)
        """
        size = min(len(frames), BBConstants.bulk_transfer_max.value)
        command = struct.pack("<BHH", SpiModeCommand.write_then_read.value, size, 0)
        self._write(command)
        self._send_command_check_response(frames, bytes([BBConstants.response_ok.value]))
        if wait_for_ack:
            try:
                # minimum delay between ack and response is 5-7ms
                time.sleep(self.ACK_WAIT_DELAY)
                self._read_frame_header()
            except AssertionError as error:
                # retry reading the SPI header frame in case check has failed
                if retry_cnt > 0:
                    logger.error(
                        f"{error} (retry {self.HDR_FRAME_RETRY_CNT-retry_cnt+1}/{self.HDR_FRAME_RETRY_CNT})"
                    )
                    retry_cnt -= 1
                    self._send_frame(frames, wait_for_ack, retry_cnt)
                else:
                    raise SPSDKError("Failed retrying reading the SPI header frame") from error

    def _read_default(self, size: int) -> bytes:
        """Read 'length' amount of bytes from BUSPAL SPI device.

        :return: Data read from the device
        """
        size = min(size, BBConstants.bulk_transfer_max.value)
        command = struct.pack("<BHH", SpiModeCommand.write_then_read.value, 0, size)
        self._send_command_check_response(command, bytes([BBConstants.response_ok.value]))
        return self._read(size)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""NXP USB Device Scanner."""

import sys
from typing import IO

import click
from click_option_group import MutuallyExclusiveOptionGroup, optgroup

from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import spsdk_apps_common_options
from spsdk.apps.utils.utils import catch_spsdk_error
from spsdk.utils import nxpdevscan


@click.command(name="nxpdevscan")
@click.option(
    "-e",
    "--extend-vids",
    multiple=True,
    default=[],
    help="VID in hex to extend search.",
)
@click.option("-o", "--out", default="-", type=click.File("w"))
# NOTE: The MutuallyExclusiveOptionGroup doesn't work for flags, we keep it just for display purposes
@optgroup.group("Narrow down the scope of scanning", cls=MutuallyExclusiveOptionGroup)
@optgroup.option(
    "-a",
    "--all",
    "scope",
    flag_value="all",
    default=True,
    help="Search for all devices (default)",
)
@optgroup.option(
    "-u",
    "--usb",
    "scope",
    flag_value="usb",
    help="Search only for USB devices",
)
@optgroup.option(
    "-sd",
    "--sdio",
    "scope",
    flag_value="sdio",
    help="Search only for SDIO devices",
)
@optgroup.option(
    "-p",
    "--port",
    "scope",
    flag_value="port",
    help="Search only for UART devices",
)
@optgroup.option(
    "-l",
    "--lpcusbsio",
    "scope",
    flag_value="lpcusbsio",
    help="Search only for USBSIO devices",
)
@spsdk_apps_common_options
def main(extend_vids: str, out: IO[str], scope: str, log_level: int) -> None:
    """Utility listing all connected NXP USB and UART devices."""
    spsdk_logger.install(level=log_level)
    additional_vids = [int(vid, 16) for vid in extend_vids]

    if scope in ["all", "sdio"]:
        nxp_sdio_devices = nxpdevscan.search_nxp_sdio_devices()
        if out.name == "<stdout>":
            click.echo(8 * "-" + " Connected NXP SDIO Devices " + 8 * "-" + "\n", out)
        for sdio_dev in nxp_sdio_devices:
            click.echo(sdio_dev.info(), out)
            click.echo("", out)

    if scope in ["all", "usb"]:
        nxp_usb_devices = nxpdevscan.search_nxp_usb_devices(additional_vids)
        if out.name == "<stdout>":
            click.echo(8 * "-" + " Connected NXP USB Devices " + 8 * "-" + "\n", out)
        for usb_dev in nxp_usb_devices:
            click.echo(usb_dev.info(), out)
            click.echo("", out)

    if scope in ["all", "port"]:
        nxp_uart_devices = nxpdevscan.search_nxp_uart_devices()
        if out.name == "<stdout>":
            click.echo(8 * "-" + " Connected NXP UART Devices " + 8 * "-" + "\n", out)
        for uart_dev in nxp_uart_devices:
            click.echo(uart_dev.info(), out)
            click.echo("", out)

    if scope in ["all", "lpcusbsio"]:
        nxp_sio_devices = nxpdevscan.search_libusbsio_devices()
        if out.name == "<stdout>":
            click.echo(8 * "-" + " Connected NXP SIO Devices " + 8 * "-" + "\n", out)
        for sio_dev in nxp_sio_devices:
            click.echo(sio_dev.info(), out)
            click.echo("", out)


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Secure Boot Image Class."""

from datetime import datetime
from typing import List, Optional, Sequence

from spsdk import SPSDKError
from spsdk.utils.crypto import crypto_backend
from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.misc import DebugInfo, align

from ..misc import BcdVersion3, BcdVersion3Format, SecBootBlckSize
from .headers import BootSectionHeaderV1, SectionHeaderItemV1, SecureBootHeaderV1
from .sections import BootSectionV1


########################################################################################################################
# Secure Boot Image Class (Version 1.x)
########################################################################################################################
class SecureBootV1(BaseClass):
    """SB file 1.x."""

    def __init__(
        self,
        version: str = "1.0",
        flags: int = 0,
        drive_tag: int = 0,
        product_version: BcdVersion3Format = BcdVersion3.DEFAULT,
        component_version: BcdVersion3Format = BcdVersion3.DEFAULT,
        dek: Optional[bytes] = None,
        mac: Optional[bytes] = None,
        digest: bytes = b"\0" * 20,
        timestamp: Optional[datetime] = None,
    ):
        """Initialize Secure Boot Image V1.x.

        :param version: string in format #.#
            Major version of the boot image format, currently 1.
            Minor version of the boot image format, currently 1 or 2.
        :param flags: for the header, 0 by default: Flags associated with the entire image.
        :param product_version: Product version.
        :param component_version: Component version.
        :param drive_tag: For header: identifier for the disk drive or partition containing this image.
        :param dek: DEK key for encrypted SB file; this is not supported yet
        :param mac: MAC for encrypted SB file, this is not supported yet
        :param digest: SHA-1 digest of all fields of the header (it will be updated before export anyway)
            The first 16 bytes (of 20 total) also act as the initialization vector for CBC-encrypted regions.
        :param timestamp: datetime of the file creation, use None for current date/time
            Fixed value should be used only for regression testing to generate same results
        """
        self._dek = dek if dek else crypto_backend().random_bytes(32)
        self._mac = mac if mac else crypto_backend().random_bytes(32)
        self._header = SecureBootHeaderV1(
            version=version,
            product_version=product_version,
            component_version=component_version,
            flags=flags,
            drive_tag=drive_tag,
            digest=digest,
            timestamp=timestamp,
        )
        self._sections_hdr_table: List[SectionHeaderItemV1] = []
        self._sections: List[BootSectionV1] = []
        self._signature = None

    def __str__(self) -> str:
        return self.info()

    @property
    def first_boot_section_id(self) -> int:
        """Return id of first boot section."""
        return self._header.first_boot_section_id

    @first_boot_section_id.setter
    def first_boot_section_id(self, value: int) -> None:
        if value > len(self._sections):
            raise SPSDKError("Invalid length of section")
        self._header.first_boot_section_id = value

    @property
    def size(self) -> int:
        """Return size of the binary representation in bytes."""
        result = self._header.size
        for sect_hdr in self._sections_hdr_table:
            result += sect_hdr.size
        for sect in self._sections:
            result += sect.size
        result += 32  # authentication
        return result

    def info(self) -> str:
        """Return text info about the instance, multi-line string."""
        result = "[SB]\n"
        result += "[SB-header]\n"
        result += self._header.info()
        result += "[Sections-Header-Table]\n"
        for sect_hdr in self._sections_hdr_table:
            result += "[Section-Header]\n"
            result += sect_hdr.info()
        result += "[Sections]\n"
        for sect in self._sections:
            result += "[Section]\n"
            result += sect.info()
        return result

    @property
    def sections(self) -> Sequence[BootSectionV1]:
        """Return sequence of all sections on the SB file."""
        return self._sections

    def append(self, section: BootSectionV1) -> None:
        """Add section into the SB file.

        :param section: to be added
        """
        assert isinstance(section, BootSectionV1)
        self._sections.append(section)

    def validate(self) -> None:
        """Validate settings.

        :raises SPSDKError: If the settings is not consistent
        """
        if not self._sections:
            raise SPSDKError("At least one section must be defined")

    def update(self) -> None:
        """Update content."""
        # update header
        self._header.section_count = len(self._sections)

        # update ROM_LAST_TAG - set only for last bootable section
        last = True
        for sect in reversed(self._sections):
            if sect.bootable:
                sect.rom_last_tag = last
                last = False

        # update section header table
        ofs_blocks = SecBootBlckSize.to_num_blocks(
            self._header.size
            + len(self._sections) * SectionHeaderItemV1.SIZE
            + BootSectionHeaderV1.SIZE
        )
        new_hdr_table: List[SectionHeaderItemV1] = []
        for sect in self._sections:
            sect_blcks = SecBootBlckSize.to_num_blocks(sect.size - BootSectionHeaderV1.SIZE)
            hdr = SectionHeaderItemV1(sect.section_id, ofs_blocks, sect_blcks, sect.flags)
            new_hdr_table.append(hdr)
            ofs_blocks += sect_blcks
        self._sections_hdr_table = new_hdr_table

        # update image size
        self._header.image_blocks = SecBootBlckSize.to_num_blocks(self.size)

    def export(
        self,
        header_padding8: Optional[bytes] = None,
        auth_padding: Optional[bytes] = None,
        dbg_info: DebugInfo = DebugInfo.disabled(),
    ) -> bytes:
        """Serialization to binary form.

        :param header_padding8: optional header padding, 8-bytes; recommended to use None to apply random value
        :param auth_padding: optional padding used after authentication; recommended to use None to apply random value
        :param dbg_info: instance allowing to debug generated output
        :return: serialize the instance into binary data
        :raises SPSDKError: Invalid section data
        :raises SPSDKError: Invalid padding length
        """
        self.update()
        self.validate()
        dbg_info.append_section("SB-FILE-1.x")
        data = self._header.export(padding8=header_padding8, dbg_info=dbg_info)
        # header table
        dbg_info.append_section("Sections-Header-Table")
        for sect_hdr in self._sections_hdr_table:
            sect_hdr_data = sect_hdr.export()
            dbg_info.append_binary_data("Section-Header-Item", sect_hdr_data)
            data += sect_hdr_data
        # sections
        dbg_info.append_section("Sections")
        for sect in self._sections:
            sect_data = sect.export(dbg_info)
            if len(sect_data) != sect.size:
                raise SPSDKError("Invalid section data")
            data += sect_data
        # authentication: SHA1
        auth_code = crypto_backend().hash(data, "sha1")
        dbg_info.append_binary_section("SHA1", auth_code)
        data += auth_code
        # padding
        padding_len = align(len(auth_code), SecBootBlckSize.BLOCK_SIZE) - len(auth_code)
        if auth_padding is None:
            auth_padding = crypto_backend().random_bytes(padding_len)
        if padding_len != len(auth_padding):
            raise SPSDKError("Invalid padding length")
        data += auth_padding
        dbg_info.append_binary_section("padding", auth_padding)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "SecureBootV1":
        """Convert binary data into the instance (deserialization).

        :param data: given binary data to be converted
        :param offset: to start parsing the data
        :return: converted instance
        :raises SPSDKError: raised when digest does not match
        :raises SPSDKError: Raised when section is invalid
        """
        obj = SecureBootV1()
        cur_pos = offset
        obj._header = SecureBootHeaderV1.parse(data, cur_pos)
        cur_pos += obj._header.size
        # sections header table
        for _ in range(obj._header.section_count):
            sect_header = SectionHeaderItemV1.parse(data, cur_pos)
            obj._sections_hdr_table.append(sect_header)
            cur_pos += sect_header.size
        # sections
        new_pos = offset + obj._header.first_boot_tag_block * SecBootBlckSize.BLOCK_SIZE
        if new_pos < cur_pos:
            raise SPSDKError("Invalid section")
        cur_pos = new_pos
        for _ in range(obj._header.section_count):
            boot_sect = BootSectionV1.parse(data, cur_pos)
            obj.append(boot_sect)
            cur_pos += boot_sect.size
        # authentication code
        sha1_auth = crypto_backend().hash(data[offset:cur_pos], "sha1")
        if sha1_auth != data[cur_pos : cur_pos + len(sha1_auth)]:
            raise SPSDKError("Authentication failure: digest does not match")
        # done
        return obj



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module to handle registers configuration."""

import os
from typing import Any, Dict, List, Optional

from spsdk import SPSDKError
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.database import Database
from spsdk.utils.misc import value_to_int


class RegConfig(Database):
    """Class that helps manage the registers configuration."""

    def get_address(self, device: str) -> int:
        """Get the area address in chip memory.

        :param device: The device name.
        :return: Base address of registers.
        """
        return value_to_int(self.devices.get_by_name(device).attributes["address"])

    def get_data_file(self, device: str, revision: str) -> str:
        """Return the full path to data file (xml).

        :param device: The device name.
        :param revision: The chip revision.
        :raises SPSDKValueError: When datafile is not defined in the database
        :return: The path to data file.
        """
        file_name = self.devices.get_by_name(device).revisions.get(revision).data_file
        if not file_name:
            raise SPSDKValueError(
                f"Datafile is not defined in database: {self.path} for device {device} and revision {revision}"
            )
        dir_path = os.path.dirname(os.path.abspath(self.path))
        return os.path.join(dir_path, file_name)

    def get_antipole_regs(self, device: Optional[str] = None) -> Dict[str, str]:
        """Return the list of inverted registers.

        :param device: The device name.
        :return: The dictionary of antipole registers.
        """
        val = self.get_value("inverted_regs", device, default={})
        assert isinstance(val, dict)
        return dict(val)

    def get_computed_fields(self, device: Optional[str] = None) -> Dict[str, Dict[str, str]]:
        """Return the list of computed fields (not used in config YML files).

        :param device: The device name, if not specified, the general value is used.
        :return: The dictionary of computed fields.
        """
        val = self.get_value("computed_fields", device, default={})
        assert isinstance(val, dict)
        return dict(val)

    def get_computed_registers(self, device: Optional[str] = None) -> Dict[str, Any]:
        """Return the dictionary of computed registers.

        :param device: The device name, if not specified, the general value is used.
        :return: The dictionary of computed registers.
        """
        val = self.get_value("computed_registers", device, default={})
        assert isinstance(val, dict)
        return val

    def get_grouped_registers(self, device: Optional[str] = None) -> List[dict]:
        """Return the list of grouped registers description.

        :param device: The device name, if not specified, the general value is used.
        :return: The list of grouped registers descriptions.
        """
        val = self.get_value("grouped_registers", device, default=[])
        assert isinstance(val, list)
        return list(val)

    def get_ignored_registers(self, device: Optional[str] = None) -> List[str]:
        """Return the list of ignored registers.

        :param device: The device name, if not specified, the general value is used.
        :return: The list of ignored register.
        """
        val = self.get_value("ignored_registers", device, default=[])
        assert isinstance(val, list)
        return list(val)

    def get_ignored_fields(self, device: Optional[str] = None) -> List[str]:
        """Return the list of ignored fields.

        :param device: The device name, if not specified, the general value is used.
        :return: The list of ignored fields.
        """
        val = self.get_value("ignored_fields", device, default=[])
        assert isinstance(val, list)
        return val

    def get_seal_start_address(self, device: Optional[str] = None) -> Optional[str]:
        """Return the seal start address.

        :param device: The device name, if not specified, the general value is used.
        :return: The seal start register name.
        :raises SPSDKError: When seal start address has invalid name
        """
        val = self.get_value("seal_start", device)
        if not (val is None or isinstance(val, str)):
            raise SPSDKError("Invalid seal start address name")
        return val

    def get_seal_count(self, device: Optional[str] = None) -> Optional[int]:
        """Return the seal count.

        :param device: The device name, if not specified, the general value is used.
        :return: The seal count.
        :raises SPSDKError: When there is invalid seal count
        """
        val = self.get_value("seal_count", device)
        if not (val is None or isinstance(val, int)):
            raise SPSDKError("Invalid seal count")
        return val

    def get_value(
        self, key: str, device: Optional[str] = None, default: Optional[Any] = None
    ) -> Any:
        """Return any parameter by key.

        :param key: The Key of the parameter to be returned.
        :param device: The device name.
        :param default: The default Value in case that is not specified in config file.
        :return: The Value of parameter by handled Key.
        """
        return self.get_device_value(key=key, device=device, default=default)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for working with Protected Flash Region."""

import os

from spsdk import SPSDK_DATA_FOLDER

PFR_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "pfr")

from .exceptions import (
    SPSDKPfrcMissingConfigError,
    SPSDKPfrConfigError,
    SPSDKPfrConfigReadError,
    SPSDKPfrError,
    SPSDKPfrRotkhIsNotPresent,
)
from .pfr import CFPA, CMPA, ROMCFG, PfrConfiguration
from .processor import Processor
from .translator import Translator



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright (c) 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for USB-SIO communication with a target device using MBoot protocol."""

import logging
import re
from typing import List, Optional

import libusbsio
from libusbsio.libusbsio import LIBUSBSIO

from spsdk import SPSDKError, SPSDKValueError
from spsdk.mboot.exceptions import McuBootConnectionError
from spsdk.utils.exceptions import SPSDKTimeoutError
from spsdk.utils.misc import value_to_int
from spsdk.utils.usbfilter import USBDeviceFilter

from .uart import Uart

logger = logging.getLogger(__name__)


def _get_usbsio() -> LIBUSBSIO:
    """Wraps getting USBSIO library to raise SPSDK errors in case of problem.

    :return: LIBUSBSIO object
    :raises SPSDKError: When libusbsio library error or if no bridge device found
    """
    try:
        # get the global singleton instance of LIBUSBSIO library
        libusbsio_logger = logging.getLogger("libusbsio")
        return libusbsio.usbsio(loglevel=libusbsio_logger.getEffectiveLevel())
    except libusbsio.LIBUSBSIO_Exception as e:
        raise SPSDKError(f"Error in libusbsio interface: {e}") from e
    except Exception as e:
        raise SPSDKError(str(e)) from e


def get_usbsio_devices(config: Optional[str] = None) -> List[int]:
    """Returns list of ports indexes of USBSIO devices.

    It could be filtered by standard SPSDK USB filters.

    :param config: Could contain USB filter configuration, defaults to None
    :return: List of port indexes of founded USBSIO device
    """

    def _filter_usb(sio: LIBUSBSIO, ports: List[int], flt: str) -> List[int]:
        """Filter the  LIBUSBSIO device.

        :param sio: LIBUSBSIO instance.
        :param ports: Input list of LIBUSBSIO available ports.
        :param flt: Filter string (PATH, PID/VID, SERIAL_NUMBER)
        :raises SPSDKError: When libusbsio library error or if no bridge device found
        :return: List with selected device, empty list otherwise.
        """
        usb_filter = USBDeviceFilter(flt.casefold())
        port_indexes = []
        for port in ports:
            info = sio.GetDeviceInfo(port)
            if not info:
                raise SPSDKError(f"Cannot retrive information from LIBUSBSIO device {port}.")
            dev_info = {
                "vendor_id": info.vendor_id,
                "product_id": info.product_id,
                "serial_number": info.serial_number,
                "path": info.path,
            }
            if usb_filter.compare(dev_info):
                port_indexes.append(port)
                break
        return port_indexes

    cfg = config.split(",") if config else []
    port_indexes = []

    sio = _get_usbsio()
    # it may already be open (?), in that case, just close it - We are scan function!
    if sio.IsOpen():
        sio.Close()

    port_indexes.extend(list(range(sio.GetNumPorts())))

    # filter out the USB devices
    if cfg and cfg[0] == "usb":
        port_indexes = _filter_usb(sio, port_indexes, cfg[1])

    return port_indexes


def scan_usbsio(config: Optional[str] = None, timeout: int = 5000) -> List["UsbSio"]:
    """Scan connected USB-SIO bridge devices.

    :param config: Configuration string identifying spi or i2c SIO interface
                    and could filter out USB devices
    :param timeout: Read timeout in milliseconds, defaults to 5000
    :return: List of matching UsbSio devices
    :raises SPSDKError: When libusbsio library error or if no bridge device found
    :raises SPSDKValueError: Invalid configuration detected.
    """
    cfg = config.split(",") if config else []
    re_spi = re.compile(r"^spi(?P<index>\d*)")
    re_i2c = re.compile(r"^i2c(?P<index>\d*)")
    spi = None
    i2c = None
    for cfg_part in cfg:
        match_i2c = re_i2c.match(cfg_part.lower())
        if match_i2c:
            i2c = value_to_int(match_i2c.group("index"), 0)
        match_spi = re_spi.match(cfg_part.lower())
        if match_spi:
            spi = value_to_int(match_spi.group("index"), 0)
    if i2c is not None and spi is not None:
        raise SPSDKValueError(f"Cannot be specified spi and i2c together in configuration: {cfg}")
    intf_specified = i2c is not None or spi is not None

    port_indexes = get_usbsio_devices(config)
    sio = _get_usbsio()
    devices: List["UsbSio"] = []
    for port in port_indexes:
        if not sio.Open(port):
            raise SPSDKError(f"Cannot open libusbsio bridge {port}.")
        i2c_ports = sio.GetNumI2CPorts()
        if i2c_ports:
            if i2c is not None:
                devices.append(UsbSioI2C(dev=port, port=i2c, config=config, timeout=timeout))
            elif not intf_specified:
                devices.extend(
                    [UsbSioI2C(dev=port, port=p, timeout=timeout) for p in range(i2c_ports)]
                )
        spi_ports = sio.GetNumSPIPorts()
        if spi_ports:
            if spi is not None:
                devices.append(UsbSioSPI(dev=port, port=spi, config=config, timeout=timeout))
            elif not intf_specified:
                devices.extend(
                    [UsbSioSPI(dev=port, port=p, timeout=timeout) for p in range(spi_ports)]
                )
        if sio.Close() < 0:
            raise SPSDKError(f"Cannot close libusbsio bridge {port}.")
        # re-init the libusb to prepare it for next open
        sio.GetNumPorts()
    return devices


def scan_usbsio_i2c(config: Optional[str] = None, timeout: int = 5000) -> List["UsbSioI2C"]:
    """Scan connected USB-SIO bridge devices and return just I2C devices.

    :param config: Configuration string identifying spi or i2c SIO interface
                    and could filter out USB devices.
    :param timeout: Read timeout in milliseconds, defaults to 5000
    :return: List of matching UsbSioI2C devices only
    """
    devices = scan_usbsio(config, timeout)
    return [x for x in devices if isinstance(x, UsbSioI2C)]


def scan_usbsio_spi(config: Optional[str] = None, timeout: int = 5000) -> List["UsbSioSPI"]:
    """Scan connected USB-SIO bridge devices and return just SPI devices.

    :param config: Configuration string identifying spi or i2c SIO interface
                    and could filter out USB devices.
    :param timeout: Read timeout in milliseconds, defaults to 5000
    :return: List of matching UsbSioSPI devices only
    """
    devices = scan_usbsio(config, timeout)
    return [x for x in devices if isinstance(x, UsbSioSPI)]


class UsbSio(Uart):
    """USBSIO general interface, base class for SPI or I2C communication over LIBUSBSIO.

    This class inherits from Uart communication as the SPI/I2C protocol is the same.
    The Uart's read and write methods are leveraged. The low-level _read and _write
    methods are overridden.
    """

    @property
    def is_opened(self) -> bool:
        """Indicates whether interface is open."""
        return bool(self.port)

    def __init__(self, dev: int = 0, config: Optional[str] = None, timeout: int = 5000) -> None:
        """Initialize the Interface object.

        :param dev: device index to be used, default is set to 0
        :param config: configuration string identifying spi or i2c SIO interface
        :param timeout: read timeout in milliseconds, defaults to 5000
        :raises SPSDKError: When LIBUSBSIO device is not opened.
        """
        # device is the LIBUSBSIO.PORT instance (LIBUSBSIO.SPI or LIBUSBSIO.I2C class)
        self.port: Optional[LIBUSBSIO.PORT] = None

        # work with the global LIBUSBSIO instance
        self.dev_ix = dev
        self.sio = _get_usbsio()

        super().__init__(timeout=timeout)

        # store USBSIO configuration and version
        self.config = config

    def info(self) -> str:
        """Return string containing information about the interface."""
        class_name = self.__class__.__name__
        config = f":'{self.config}'" if self.config else ""
        return f"libusbsio interface ({class_name}){config}"

    @staticmethod
    def get_interface_cfg(config: str, interface: str) -> str:
        """Return part of interface config.

        :param config: Full config of LIBUSBSIO
        :param interface: Name of interface to find.
        :return: Part with interface config.
        """
        i = config.rfind(interface)
        if i < 0:
            return ""
        return config[i]


class UsbSioSPI(UsbSio):
    """USBSIO SPI interface."""

    FRAME_START_BYTE_NOT_READY = 0xFF

    def __init__(
        self,
        config: Optional[str] = None,
        dev: int = 0,
        port: int = 0,
        ssel_port: int = 0,
        ssel_pin: int = 15,
        speed_khz: int = 1000,
        cpol: int = 1,
        cpha: int = 1,
        timeout: int = 5000,
    ) -> None:
        """Initialize the UsbSioSPI Interface object.

        :param config: configuration string passed from command line
        :param dev: device index to be used, default is set to 0
        :param port: default SPI port to be used, typically 0 as only one port is supported by LPCLink2/MCULink
        :param ssel_port: bridge GPIO port used to drive SPI SSEL signal
        :param ssel_pin: bridge GPIO pin used to drive SPI SSEL signal
        :param speed_khz: SPI clock speed in kHz
        :param cpol: SPI clock polarity mode
        :param cpha: SPI clock phase mode
        :param timeout: read timeout in milliseconds, defaults to 5000
        :raises SPSDKError: When port configuration cannot be parsed
        """
        super().__init__(dev=dev, config=config, timeout=timeout)

        # default configuration taken from parameters (and their default values)
        self.spi_port = port
        self.spi_sselport = ssel_port
        self.spi_sselpin = ssel_pin
        self.spi_speed_khz = speed_khz
        self.spi_cpol = cpol
        self.spi_cpha = cpha

        # values can be also overridden by a configuration string
        if config:
            # config format: spi[,<port>,<pin>,<speed>,<cpol>,<cpha>]
            cfg = self.get_interface_cfg(config, "spi").split(",")
            try:
                self.spi_sselport = int(cfg[1], 0)
                self.spi_sselpin = int(cfg[2], 0)
                self.spi_speed_khz = int(cfg[3], 0)
                self.spi_cpol = int(cfg[4], 0)
                self.spi_cpha = int(cfg[5], 0)
            except IndexError:
                pass
            except Exception as e:
                raise SPSDKError(
                    "Cannot parse lpcusbsio SPI parameters.\n"
                    "Expected: spi[,<port>,<pin>,<speed_kHz>,<cpol>,<cpha>]\n"
                    f"Given:    {config}"
                ) from e

    def open(self) -> None:
        """Open the interface."""
        if not self.sio.IsOpen():
            self.sio.Open(self.dev_ix)

        self.port: LIBUSBSIO.SPI = self.sio.SPI_Open(
            portNum=self.spi_port,
            busSpeed=self.spi_speed_khz * 1000,
            cpol=self.spi_cpol,
            cpha=self.spi_cpha,
        )
        if not self.port:
            raise SPSDKError("Cannot open lpcusbsio SPI interface.\n")

    def close(self) -> None:
        """Close the interface."""
        if self.port:
            self.port.Close()
            self.port = None
            # if not self.sio.IsAnyPortOpen(): #TODO temporary hack till new version of libusbsio greater than 2.1.11
            self.sio.Close()

    def _read(self, length: int) -> bytes:
        """Read 'length' amount for bytes from device.

        :param length: Number of bytes to read
        :return: Data read from the device
        :raises McuBootConnectionError: When reading data from device fails
        :raises TimeoutError: When no data received
        """
        try:
            (data, result) = self.port.Transfer(
                devSelectPort=self.spi_sselport,
                devSelectPin=self.spi_sselpin,
                txData=None,
                size=length,
            )
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e
        if result < 0 or not data:
            raise SPSDKTimeoutError()
        logger.debug(f"<{' '.join(f'{b:02x}' for b in data)}>")
        return data

    def _write(self, data: bytes) -> None:
        """Send data to device.

        :param data: Data to send
        :raises McuBootConnectionError: When sending the data fails
        :raises SPSDKTimeoutError: When data could not be written
        """
        logger.debug(f"[{' '.join(f'{b:02x}' for b in data)}]")
        try:
            (dummy, result) = self.port.Transfer(
                devSelectPort=self.spi_sselport, devSelectPin=self.spi_sselpin, txData=data
            )
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e
        if result < 0:
            raise SPSDKTimeoutError()


class UsbSioI2C(UsbSio):
    """USBSIO I2C interface."""

    def __init__(
        self,
        config: Optional[str] = None,
        dev: int = 0,
        port: int = 0,
        address: int = 0x10,
        speed_khz: int = 100,
        timeout: int = 5000,
    ) -> None:
        """Initialize the UsbSioI2C Interface object.

        :param config: configuration string passed from command line
        :param dev: device index to be used, default is set to 0
        :param port: default I2C port to be used, typically 0 as only one port is supported by LPCLink2/MCULink
        :param address: I2C target device address
        :param speed_khz: I2C clock speed in kHz
        :param timeout: read timeout in milliseconds, defaults to 5000
        :raises SPSDKError: When port configuration cannot be parsed
        """
        super().__init__(dev=dev, config=config, timeout=timeout)

        # default configuration taken from parameters (and their default values)
        self.i2c_port = port
        self.i2c_address = address
        self.i2c_speed_khz = speed_khz

        # values can be also overridden by a configuration string
        if config:
            # config format: i2c[,<address>,<speed>]
            cfg = self.get_interface_cfg(config, "i2c").split(",")
            try:
                self.i2c_address = int(cfg[1], 0)
                self.i2c_speed_khz = int(cfg[2], 0)
            except IndexError:
                pass
            except Exception as e:
                raise SPSDKError(
                    "Cannot parse lpcusbsio I2C parameters.\n"
                    "Expected: i2c[,<address>,<speed_kHz>]\n"
                    f"Given:    {config}"
                ) from e

    def open(self) -> None:
        """Open the interface."""
        if not self.sio.IsOpen():
            self.sio.Open(self.dev_ix)

        self.port: LIBUSBSIO.I2C = self.sio.I2C_Open(
            clockRate=self.i2c_speed_khz * 1000, portNum=self.i2c_port
        )
        if not self.port:
            raise SPSDKError("Cannot open lpcusbsio I2C interface.\n")

    def close(self) -> None:
        """Close the interface."""
        if self.port:
            # Preventive Reset of I2C device to be ready to reopen in case of any pending problem
            self.port.Reset()
            self.port.Close()
            self.port = None
            # if not self.sio.IsAnyPortOpen(): #TODO temporary hack till new version of libusbsio greater than 2.1.11
            self.sio.Close()

    def _read(self, length: int) -> bytes:
        """Read 'length' amount for bytes from device.

        :param length: Number of bytes to read
        :return: Data read from the device
        :raises SPSDKTimeoutError: Time-out
        :raises McuBootConnectionError: When reading data from device fails
        :raises SPSDKTimeoutError: When no data received
        """
        try:
            (data, result) = self.port.DeviceRead(devAddr=self.i2c_address, rxSize=length)
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e
        if result < 0 or not data:
            raise SPSDKTimeoutError()
        logger.debug(f"<{' '.join(f'{b:02x}' for b in data)}>")
        return data

    def _write(self, data: bytes) -> None:
        """Send data to device.

        :param data: Data to send
        :raises McuBootConnectionError: When sending the data fails
        :raises TimeoutError: When data NAKed or could not be written
        """
        logger.debug(f"[{' '.join(f'{b:02x}' for b in data)}]")
        try:
            result = self.port.DeviceWrite(devAddr=self.i2c_address, txData=data)
        except Exception as e:
            raise McuBootConnectionError(str(e)) from e
        if result < 0:
            raise SPSDKTimeoutError()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""This module contains FCB (Flash Configuration Block) related code."""


import datetime
import os
from copy import deepcopy
from typing import Any, Dict, List

from ruamel.yaml.comments import CommentedMap as CM

from spsdk import version as spsdk_version
from spsdk.exceptions import SPSDKError, SPSDKValueError
from spsdk.image.fcb import FCB_DATA_FOLDER, FCB_DATABASE_FILE, FCB_SCH_FILE
from spsdk.image.segments_base import SegmentBase
from spsdk.utils.database import Database
from spsdk.utils.registers import Registers
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas


class FCB(SegmentBase):
    """FCB (Flash Configuration Block)."""

    def __init__(self, family: str, mem_type: str, revision: str = "latest") -> None:
        """FCB Constructor.

        :param family: Chip family.
        :param mem_type: Used memory type.
        :param revision: Optional Chip family revision.
        :raises SPSDKValueError: Unsupported family.
        """
        super().__init__(family, revision)
        mem_types = FCB.get_supported_memory_types(self.family, self.revision)
        if mem_type not in mem_types:
            raise SPSDKValueError(f"Unsupported memory type:{mem_type} not in {mem_types}")
        self.mem_type = mem_type
        self._registers = Registers(family, base_endianness="little")

        block_file_name = self.get_memory_types(self.family, self.revision)[self.mem_type]
        self._registers.load_registers_from_xml(os.path.join(FCB_DATA_FOLDER, block_file_name))

    @staticmethod
    def get_database() -> Database:
        """Get the devices database."""
        return Database(FCB_DATABASE_FILE)

    @property
    def registers(self) -> Registers:
        """Registers of segment."""
        return self._registers

    @staticmethod
    def load_from_config(config: Dict) -> "FCB":
        """Load configuration file of FCB.

        :param config: FCB configuration file.
        :return: FCB object.
        """
        family = config.get("family", "Unknown")
        mem_type = config.get("type", "Unknown")
        revision = config.get("revision", "latest")
        fcb = FCB(family=family, mem_type=mem_type, revision=revision)
        fcb_settings = config.get("fcb_settings", {})
        fcb.registers.load_yml_config(fcb_settings)
        return fcb

    def create_config(self) -> str:
        """Create current configuration YAML.

        :return: Configuration of FCB Block.
        """
        config = CM()
        config["family"] = self.family
        config["revision"] = self.revision
        config["type"] = self.mem_type
        config["fcb_settings"] = self.registers.create_yml_config()
        config.yaml_set_start_comment(
            f"FCB configuration for {self.family}.\n"
            f"Created: {datetime.datetime.now().strftime('%d/%m/%Y %H:%M:%S')}.\n"
            f"NXP SPSDK version: {spsdk_version}"
        )
        return ConfigTemplate.convert_cm_to_yaml(config)

    @classmethod
    def get_validation_schemas(
        cls, family: str, mem_type: str, revision: str = "latest"
    ) -> List[Dict[str, Any]]:
        """Create the validation schema.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param revision: Chip revision specification, as default, latest is used.
        :raises SPSDKError: Family or revision is not supported.
        :return: List of validation schemas.
        """
        fcb_obj = FCB(family, mem_type, revision)
        sch_cfg = deepcopy(ValidationSchemas.get_schema_file(FCB_SCH_FILE))
        try:
            sch_cfg["fcb_family_rev"]["properties"]["family"]["enum"] = FCB.get_supported_families()
            revisions = ["latest"]
            revisions.extend(
                fcb_obj.get_database().devices.get_by_name(family).revisions.revision_names
            )
            sch_cfg["fcb_family_rev"]["properties"]["revision"]["enum"] = revisions
            sch_cfg["fcb_family_rev"]["properties"]["type"]["enum"] = list(
                fcb_obj.get_supported_memory_types(fcb_obj.family, revision)
            )
            sch_cfg["fcb"]["properties"]["fcb_settings"] = fcb_obj.registers.get_validation_schema()
            return [sch_cfg["fcb_family_rev"], sch_cfg["fcb"]]
        except (KeyError, SPSDKError) as exc:
            raise SPSDKError(f"Family {family} or {revision} is not supported") from exc

    @staticmethod
    def get_validation_schemas_family() -> List[Dict[str, Any]]:
        """Create the validation schema just for supported families.

        :return: List of validation schemas for FCB supported families.
        """
        sch_cfg = deepcopy(ValidationSchemas.get_schema_file(FCB_SCH_FILE))
        sch_cfg["fcb_family_rev"]["properties"]["family"]["enum"] = FCB.get_supported_families()
        return [sch_cfg["fcb_family_rev"]]

    @staticmethod
    def generate_config_template(family: str, mem_type: str, revision: str = "latest") -> str:
        """Generate configuration for selected family.

        :param family: Family description.
        :param mem_type: Used memory type.
        :param revision: Chip revision specification, as default, latest is used.
        :return: Template of FCB Block.
        """
        ret = ""

        if family in FCB.get_supported_families():
            schemas = FCB.get_validation_schemas(family, mem_type, revision)
            override = {}
            override["family"] = family
            override["revision"] = revision
            override["type"] = mem_type

            ret = ConfigTemplate(
                f"Flash Configuration Block template for {family}.",
                schemas,
                override,
            ).export_to_yaml()

        return ret



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Commands used by SBFile module."""
import math
from abc import abstractmethod
from struct import calcsize, pack, unpack_from
from typing import Any, Mapping, Optional, Type

from crcmod.predefined import mkPredefinedCrcFun

from spsdk import SPSDKError
from spsdk.mboot import ExtMemId
from spsdk.sbfile.misc import SecBootBlckSize
from spsdk.utils.crypto.abstract import BaseClass
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import DebugInfo

########################################################################################################################
# Constants
########################################################################################################################

DEVICE_ID_MASK = 0xFF
DEVICE_ID_SHIFT = 0
GROUP_ID_MASK = 0xF00
GROUP_ID_SHIFT = 8


########################################################################################################################
# Enums
########################################################################################################################
class EnumCmdTag(Enum):
    """Command tags."""

    NOP = 0x0
    TAG = 0x1
    LOAD = 0x2
    FILL = 0x3
    JUMP = 0x4
    CALL = 0x5
    ERASE = 0x7
    RESET = 0x8
    MEM_ENABLE = 0x9
    PROG = 0xA
    FW_VERSION_CHECK = (0xB, "Check FW version fuse value")
    WR_KEYSTORE_TO_NV = (0xC, "Restore key-store restore to non-volatile memory")
    WR_KEYSTORE_FROM_NV = (0xD, "Backup key-store from non-volatile memory")


class EnumSectionFlag(Enum):
    """Section flags."""

    BOOTABLE = 0x0001
    CLEARTEXT = 0x0002
    LAST_SECT = 0x8000


########################################################################################################################
# Header Class
########################################################################################################################
class CmdHeader:
    """SBFile command header."""

    FORMAT = "<2BH3L"
    SIZE = calcsize(FORMAT)

    @property
    def crc(self) -> int:
        """Calculate CRC for the header data."""
        raw_data = self._raw_data(crc=0)
        checksum = 0x5A
        for i in range(1, self.SIZE):
            checksum = (checksum + raw_data[i]) & 0xFF
        return checksum

    def __init__(self, tag: int, flags: int = 0) -> None:
        """Initialize header."""
        if tag not in EnumCmdTag.tags():
            raise SPSDKError("Incorrect command tag")
        self.tag = tag
        self.flags = flags
        self.address = 0
        self.count = 0
        self.data = 0

    def __eq__(self, obj: Any) -> bool:
        return isinstance(obj, self.__class__) and vars(obj) == vars(self)

    def __ne__(self, obj: Any) -> bool:
        return not self.__eq__(obj)

    def __str__(self) -> str:
        tag = EnumCmdTag.get(self.tag, f"0x{self.tag:02X}")
        return (
            f"tag={tag}, flags=0x{self.flags:04X}, "
            f"address=0x{self.address:08X}, count=0x{self.count:08X}, data=0x{self.data:08X}"
        )

    def _raw_data(self, crc: int) -> bytes:
        """Return raw data of the header with specified CRC.

        :param crc: value to be used
        :return: binary representation of the header
        """
        return pack(self.FORMAT, crc, self.tag, self.flags, self.address, self.count, self.data)

    def export(self) -> bytes:
        """Export command header as bytes."""
        return self._raw_data(self.crc)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdHeader":
        """Parse command header from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: CMDHeader object
        :raises SPSDKError: raised when size is incorrect
        :raises SPSDKError: Raised when CRC is incorrect
        """
        if calcsize(cls.FORMAT) > len(data) - offset:
            raise SPSDKError("Incorrect size")
        obj = cls(EnumCmdTag.NOP)
        (crc, obj.tag, obj.flags, obj.address, obj.count, obj.data) = unpack_from(
            cls.FORMAT, data, offset
        )
        if crc != obj.crc:
            raise SPSDKError("CRC does not match")
        return obj


########################################################################################################################
# Commands Classes
########################################################################################################################
class CmdBaseClass(BaseClass):
    """Base class for all commands."""

    # bit mask for device ID inside flags
    ROM_MEM_DEVICE_ID_MASK = 0xFF00
    # shift for device ID inside flags
    ROM_MEM_DEVICE_ID_SHIFT = 8
    # bit mask for group ID inside flags
    ROM_MEM_GROUP_ID_MASK = 0xF0
    # shift for group ID inside flags
    ROM_MEM_GROUP_ID_SHIFT = 4

    def __init__(self, tag: int) -> None:
        """Initialize CmdBase."""
        self._header = CmdHeader(tag)

    @property
    def header(self) -> CmdHeader:
        """Return command header."""
        return self._header

    @property
    def raw_size(self) -> int:
        """Return size of the command in binary format (including header)."""
        return CmdHeader.SIZE  # this is default implementation

    def __str__(self) -> str:
        return "Command: " + str(self._header)  # default implementation: use command name

    def info(self) -> str:
        """Return text info about the instance."""
        return str(self) + "\n"  # default implementation is same as __str__

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Return object serialized into bytes."""
        dbg_info.append_section("Command:" + EnumCmdTag.name(self.header.tag))
        cmd_data = self._header.export()  # default implementation
        dbg_info.append_binary_data("cmd-header", cmd_data)
        return cmd_data

    @classmethod
    @abstractmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdBaseClass":
        """Deserialize object from binary."""


class CmdNop(CmdBaseClass):
    """Command NOP class."""

    def __init__(self) -> None:
        """Initialize Command Nop."""
        super().__init__(EnumCmdTag.NOP)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdNop":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: CMD Nop object
        :raises SPSDKError: When there is incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.NOP:
            raise SPSDKError("Incorrect header tag")
        return cls()


class CmdTag(CmdBaseClass):
    """Command TAG class.

    It is also used as header for boot section for SB file 1.x.
    """

    def __init__(self) -> None:
        """Initialize Command Tag."""
        super().__init__(EnumCmdTag.TAG)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdTag":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: parsed instance
        :raises SPSDKError: When there is incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.TAG:
            raise SPSDKError("Incorrect header tag")
        result = cls()
        result._header = header
        return result


class CmdLoad(CmdBaseClass):
    """Command Load. The load statement is used to store data into the memory."""

    @property
    def address(self) -> int:
        """Return address in target processor to load data."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Setter.

        :param value: address in target processor to load data
        :raises SPSDKError: When there is incorrect address
        """
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect address")
        self._header.address = value

    @property
    def flags(self) -> int:
        """Return command's flag."""
        return self._header.flags

    @flags.setter
    def flags(self, value: int) -> None:
        """Set command's flag."""
        self._header.flags = value

    @property
    def raw_size(self) -> int:
        """Return aligned size of the command including header and data."""
        size = CmdHeader.SIZE + len(self.data)
        if size % CmdHeader.SIZE:
            size += CmdHeader.SIZE - (size % CmdHeader.SIZE)
        return size

    def __init__(self, address: int, data: bytes, mem_id: int = 0) -> None:
        """Initialize CMD Load."""
        super().__init__(EnumCmdTag.LOAD)
        assert isinstance(data, (bytes, bytearray))
        self.address = address
        self.data = bytes(data)
        self.mem_id = mem_id

        device_id = get_device_id(mem_id)
        group_id = get_group_id(mem_id)

        self.flags |= (self.flags & ~self.ROM_MEM_DEVICE_ID_MASK) | (
            (device_id << self.ROM_MEM_DEVICE_ID_SHIFT) & self.ROM_MEM_DEVICE_ID_MASK
        )

        self.flags |= (self.flags & ~self.ROM_MEM_GROUP_ID_MASK) | (
            (group_id << self.ROM_MEM_GROUP_ID_SHIFT) & self.ROM_MEM_GROUP_ID_MASK
        )

    def __str__(self) -> str:
        return (
            f"LOAD: Address=0x{self.address:08X}, DataLen={len(self.data)}, "
            f"Flags=0x{self.flags:08X}, MemId=0x{self.mem_id:08X}"
        )

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Export command as binary."""
        self._update_data()
        result = super().export(dbg_info)
        dbg_info.append_binary_section("load-data", self.data)
        return result + self.data

    def _update_data(self) -> None:
        """Update command data."""
        # padding data
        self.data = SecBootBlckSize.align_block_fill_random(self.data)
        # update header
        self._header.count = len(self.data)
        crc32_function = mkPredefinedCrcFun("crc-32-mpeg")
        self._header.data = crc32_function(self.data, 0xFFFFFFFF)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdLoad":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: CMD Load object
        :raises SPSDKError: Raised when there is invalid CRC
        :raises SPSDKError: When there is incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.LOAD:
            raise SPSDKError("Incorrect header tag")
        offset += CmdHeader.SIZE
        header_count = SecBootBlckSize.align(header.count)
        cmd_data = data[offset : offset + header_count]
        crc32_function = mkPredefinedCrcFun("crc-32-mpeg")
        if header.data != crc32_function(cmd_data, 0xFFFFFFFF):
            raise SPSDKError("Invalid CRC in the command header")
        device_id = (header.flags & cls.ROM_MEM_DEVICE_ID_MASK) >> cls.ROM_MEM_DEVICE_ID_SHIFT
        group_id = (header.flags & cls.ROM_MEM_GROUP_ID_MASK) >> cls.ROM_MEM_GROUP_ID_SHIFT
        mem_id = get_memory_id(device_id, group_id)
        obj = CmdLoad(header.address, cmd_data, mem_id)
        obj.header.data = header.data
        obj.header.flags = header.flags
        obj._update_data()
        return obj


class CmdFill(CmdBaseClass):
    """Command Fill class."""

    PADDING_VALUE = 0x00

    @property
    def address(self) -> int:
        """Return address of the command Fill."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Set address for the command Fill."""
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect address")
        self._header.address = value

    @property
    def raw_size(self) -> int:
        """Calculate raw size of header."""
        size = CmdHeader.SIZE
        size += len(self._pattern) - 4
        if size % CmdHeader.SIZE:
            size += CmdHeader.SIZE - (size % CmdHeader.SIZE)
        return size

    def __init__(self, address: int, pattern: int, length: Optional[int] = None) -> None:
        """Initialize Command Fill.

        :param address: to write data
        :param pattern: data to be written
        :param length: length of data to be filled, defaults to 4
        :raises SPSDKError: Raised when size is not aligned to 4 bytes
        """
        super().__init__(EnumCmdTag.FILL)
        length = length or 4
        if length % 4:
            raise SPSDKError("Length of memory range to fill must be a multiple of 4")
        # if the pattern is a zero, the length is considered also as zero and the
        # conversion to bytes produces empty byte "array", which is wrong, as
        # zero should be converted to zero byte. Thus in case the pattern_len
        # evaluates to 0, we set it to 1.
        pattern_len = pattern.bit_length() / 8 or 1
        # We can get a number of 3 bytes, so we consider this as a word and set
        # the length to 4 bytes with the first byte being zero.
        if 3 == math.ceil(pattern_len):
            pattern_len = 4
        pattern_bytes = pattern.to_bytes(math.ceil(pattern_len), "big")
        # The pattern length is computed above, but as we transform the number
        # into bytes, compute the len again just in case - a bit paranoid
        # approach chosen.
        if len(pattern_bytes) not in [1, 2, 4]:
            raise SPSDKError("Pattern must be 1, 2 or 4 bytes long")
        replicate = 4 // len(pattern_bytes)
        final_pattern = replicate * pattern_bytes
        self.address = address
        self._pattern = final_pattern
        # update header
        self._header.data = unpack_from(">L", self._pattern)[0]
        self._header.count = length

    @property
    def pattern(self) -> bytes:
        """Return binary data to fill."""
        return self._pattern

    def __str__(self) -> str:
        return f"FILL: Address=0x{self.address:08X}, Pattern=" + " ".join(
            f"{byte:02X}" for byte in self._pattern
        )

    def export(self, dbg_info: DebugInfo = DebugInfo.disabled()) -> bytes:
        """Return command in binary form (serialization)."""
        # export cmd
        data = super().export(dbg_info)
        # export additional data
        data = SecBootBlckSize.align_block_fill_random(data)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdFill":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: Command Fill object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.FILL:
            raise SPSDKError("Incorrect header tag")
        # The last 4 bytes of header are part of pattern value
        offset += CmdHeader.SIZE - 4
        return cls(header.address, header.data, header.count)


class CmdJump(CmdBaseClass):
    """Command Jump class."""

    @property
    def address(self) -> int:
        """Return address of the command Jump."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Set address of the command Jump."""
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect address")
        self._header.address = value

    @property
    def argument(self) -> int:
        """Return command's argument."""
        return self._header.data

    @argument.setter
    def argument(self, value: int) -> None:
        """Set command's argument."""
        self._header.data = value

    @property
    def spreg(self) -> Optional[int]:
        """Return command's Stack Pointer."""
        if self._header.flags == 2:
            return self._header.count

        return None

    @spreg.setter
    def spreg(self, value: Optional[int] = None) -> None:
        """Set command's Stack Pointer."""
        if value is None:
            self._header.flags = 0
            self._header.count = 0
        else:
            self._header.flags = 2
            self._header.count = value

    def __init__(self, address: int = 0, argument: int = 0, spreg: Optional[int] = None) -> None:
        """Initialize Command Jump."""
        super().__init__(EnumCmdTag.JUMP)
        self.address = address
        self.argument = argument
        self.spreg = spreg

    def __str__(self) -> str:
        nfo = f"JUMP: Address=0x{self.address:08X}, Argument=0x{self.argument:08X}"
        if self.spreg is not None:
            nfo += f", SP=0x{self.spreg:08X}"
        return nfo

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdJump":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: Command Jump object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.JUMP:
            raise SPSDKError("Incorrect header tag")
        return cls(header.address, header.data, header.count if header.flags else None)


class CmdCall(CmdBaseClass):
    """Command Call.

    The call statement is used for inserting a bootloader command that executes a function
    from one of the files that are loaded into the memory.
    """

    @property
    def address(self) -> int:
        """Return command's address."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Set command's address."""
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect address")
        self._header.address = value

    @property
    def argument(self) -> int:
        """Return command's argument."""
        return self._header.data

    @argument.setter
    def argument(self, value: int) -> None:
        """Set command's argument."""
        self._header.data = value

    def __init__(self, address: int = 0, argument: int = 0) -> None:
        """Initialize Command Call."""
        super().__init__(EnumCmdTag.CALL)
        self.address = address
        self.argument = argument

    def __str__(self) -> str:
        return f"CALL: Address=0x{self.address:08X}, Argument=0x{self.argument:08X}"

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdCall":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: Command Call object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.CALL:
            raise SPSDKError("Incorrect header tag")
        return cls(header.address, header.data)


class CmdErase(CmdBaseClass):
    """Command Erase class."""

    @property
    def address(self) -> int:
        """Return command's address."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Set command's address."""
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect address")
        self._header.address = value

    @property
    def length(self) -> int:
        """Return command's count."""
        return self._header.count

    @length.setter
    def length(self, value: int) -> None:
        """Set command's count."""
        self._header.count = value

    @property
    def flags(self) -> int:
        """Return command's flag."""
        return self._header.flags

    @flags.setter
    def flags(self, value: int) -> None:
        """Set command's flag."""
        self._header.flags = value

    def __init__(self, address: int = 0, length: int = 0, flags: int = 0, mem_id: int = 0) -> None:
        """Initialize Command Erase."""
        super().__init__(EnumCmdTag.ERASE)
        self.address = address
        self.length = length
        self.flags = flags
        self.mem_id = mem_id

        device_id = get_device_id(mem_id)
        group_id = get_group_id(mem_id)

        self.flags |= (self.flags & ~self.ROM_MEM_DEVICE_ID_MASK) | (
            (device_id << self.ROM_MEM_DEVICE_ID_SHIFT) & self.ROM_MEM_DEVICE_ID_MASK
        )

        self.flags |= (self.flags & ~self.ROM_MEM_GROUP_ID_MASK) | (
            (group_id << self.ROM_MEM_GROUP_ID_SHIFT) & self.ROM_MEM_GROUP_ID_MASK
        )

    def __str__(self) -> str:
        return (
            f"ERASE: Address=0x{self.address:08X}, Length={self.length}, Flags=0x{self.flags:08X}, "
            f"MemId=0x{self.mem_id:08X}"
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdErase":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: Command Erase object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.ERASE:
            raise SPSDKError("Invalid header tag")
        device_id = (header.flags & cls.ROM_MEM_DEVICE_ID_MASK) >> cls.ROM_MEM_DEVICE_ID_SHIFT
        group_id = (header.flags & cls.ROM_MEM_GROUP_ID_MASK) >> cls.ROM_MEM_GROUP_ID_SHIFT
        mem_id = get_memory_id(device_id, group_id)
        return cls(header.address, header.count, header.flags, mem_id)


class CmdReset(CmdBaseClass):
    """Command Reset class."""

    def __init__(self) -> None:
        """Initialize Command Reset."""
        super().__init__(EnumCmdTag.RESET)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdReset":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: Cmd Reset object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.RESET:
            raise SPSDKError("Invalid header tag")
        return cls()


class CmdMemEnable(CmdBaseClass):
    """Command to configure certain memory."""

    @property
    def address(self) -> int:
        """Return command's address."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Set command's address."""
        self._header.address = value

    @property
    def size(self) -> int:
        """Return command's size."""
        return self._header.count

    @size.setter
    def size(self, value: int) -> None:
        """Set command's size."""
        self._header.count = value

    @property
    def flags(self) -> int:
        """Return command's flag."""
        return self._header.flags

    @flags.setter
    def flags(self, value: int) -> None:
        """Set command's flag."""
        self._header.flags = value

    def __init__(self, address: int, size: int, mem_id: int):
        """Initialize CmdMemEnable.

        :param address: source address with configuration data for memory initialization
        :param size: size of configuration data used for memory initialization
        :param mem_id: identification of memory
        """
        super().__init__(EnumCmdTag.MEM_ENABLE)
        self.address = address
        self.mem_id = mem_id
        self.size = size

        device_id = get_device_id(mem_id)
        group_id = get_group_id(mem_id)

        self.flags |= (self.flags & ~self.ROM_MEM_DEVICE_ID_MASK) | (
            (device_id << self.ROM_MEM_DEVICE_ID_SHIFT) & self.ROM_MEM_DEVICE_ID_MASK
        )

        self.flags |= (self.flags & ~self.ROM_MEM_GROUP_ID_MASK) | (
            (group_id << self.ROM_MEM_GROUP_ID_SHIFT) & self.ROM_MEM_GROUP_ID_MASK
        )

    def __str__(self) -> str:
        return (
            f"MEM-ENABLE: Address=0x{self.address:08X}, Size={self.size}, "
            f"Flags=0x{self.flags:08X}, MemId=0x{self.mem_id:08X}"
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdMemEnable":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: Command Memory Enable object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.MEM_ENABLE:
            raise SPSDKError("Invalid header tag")
        device_id = (header.flags & cls.ROM_MEM_DEVICE_ID_MASK) >> cls.ROM_MEM_DEVICE_ID_SHIFT
        group_id = (header.flags & cls.ROM_MEM_GROUP_ID_MASK) >> cls.ROM_MEM_GROUP_ID_SHIFT
        mem_id = get_memory_id(device_id, group_id)
        return cls(header.address, header.count, mem_id)


class CmdProg(CmdBaseClass):
    """Command Program class."""

    @property
    def address(self) -> int:
        """Return address in target processor to program data."""
        return self._header.address

    @address.setter
    def address(self, value: int) -> None:
        """Setter.

        :param value: address in target processor to load data
        :raises SPSDKError: When there is incorrect address
        """
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect address")
        self._header.address = value

    @property
    def flags(self) -> int:
        """Return command's flag."""
        return self._header.flags

    @flags.setter
    def flags(self, value: int) -> None:
        """Set command's flag."""
        self._header.flags = self.is_eight_byte
        self._header.flags |= value

    @property
    def data_word1(self) -> int:
        """Return data word 1."""
        return self._header.count

    @data_word1.setter
    def data_word1(self, value: int) -> None:
        """Setter.

        :param value: first data word
        :raises SPSDKError: When there is incorrect value
        """
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect data word 1")
        self._header.count = value

    @property
    def data_word2(self) -> int:
        """Return data word 2."""
        return self._header.data

    @data_word2.setter
    def data_word2(self, value: int) -> None:
        """Setter.

        :param value: second data word
        :raises SPSDKError: When there is incorrect value
        """
        if value < 0x00000000 or value > 0xFFFFFFFF:
            raise SPSDKError("Incorrect data word 2")
        self._header.data = value

    def __init__(
        self, address: int, mem_id: int, data_word1: int, data_word2: int = 0, flags: int = 0
    ) -> None:
        """Initialize CMD Prog."""
        super().__init__(EnumCmdTag.PROG)

        if data_word2:
            self.is_eight_byte = 1
        else:
            self.is_eight_byte = 0

        if mem_id < 0 or mem_id > 0xFF:
            raise SPSDKError("Invalid ID of memory")

        self.address = address
        self.data_word1 = data_word1
        self.data_word2 = data_word2
        self.mem_id = mem_id
        self.flags = flags

        self.flags = (self.flags & ~self.ROM_MEM_DEVICE_ID_MASK) | (
            (self.mem_id << self.ROM_MEM_DEVICE_ID_SHIFT) & self.ROM_MEM_DEVICE_ID_MASK
        )

    def __str__(self) -> str:
        return (
            f"PROG: Index=0x{self.address:08X}, DataWord1=0x{self.data_word1:08X}, "
            f"DataWord2=0x{self.data_word2:08X}, Flags=0x{self.flags:08X}, MemId=0x{self.mem_id:08X}"
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdProg":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: parsed command object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.PROG:
            raise SPSDKError("Invalid header tag")
        mem_id = (header.flags & cls.ROM_MEM_DEVICE_ID_MASK) >> cls.ROM_MEM_DEVICE_ID_SHIFT
        return cls(header.address, mem_id, header.count, header.data, header.flags)


class VersionCheckType(Enum):
    """Select type of the version check: either secure or non-secure firmware to be checked."""

    SECURE_VERSION = 0
    NON_SECURE_VERSION = 1


class CmdVersionCheck(CmdBaseClass):
    """FW Version Check command class.

    Validates version of secure or non-secure firmware.
    The command fails if version is < expected.
    """

    def __init__(self, ver_type: VersionCheckType, version: int) -> None:
        """Initialize CmdVersionCheck.

        :param ver_type: version check type, see `VersionCheckType` enum
        :param version: to be checked
        :raises SPSDKError: If invalid version check type
        """
        super().__init__(EnumCmdTag.FW_VERSION_CHECK)
        if ver_type not in VersionCheckType.tags():
            raise SPSDKError("Invalid version check type")
        self.header.address = ver_type
        self.header.count = version

    @property
    def type(self) -> VersionCheckType:
        """Return type of the check version, see VersionCheckType enumeration."""
        return VersionCheckType.from_int(self.header.address)

    @property
    def version(self) -> int:
        """Return minimal version expected."""
        return self.header.count

    def __str__(self) -> str:
        return (
            f"CVER: Type={VersionCheckType.name(self.type)}, Version={str(self.version)}, "
            f"Flags=0x{self.header.flags:08X}"
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdVersionCheck":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: parsed command object
        :raises SPSDKError: If incorrect header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != EnumCmdTag.FW_VERSION_CHECK:
            raise SPSDKError("Invalid header tag")
        ver_type = VersionCheckType.from_int(header.address)
        version = header.count
        return CmdVersionCheck(ver_type, version)


class CmdKeyStoreBackupRestore(CmdBaseClass):
    """Shared, abstract implementation for key-store backup and restore command."""

    # bit mask for controller ID inside flags
    ROM_MEM_DEVICE_ID_MASK = 0xFF00
    # shift for controller ID inside flags
    ROM_MEM_DEVICE_ID_SHIFT = 8

    @classmethod
    @abstractmethod
    def cmd_id(cls) -> EnumCmdTag:
        """Return command ID.

        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")

    def __init__(self, address: int, controller_id: ExtMemId):
        """Initialize CmdKeyStoreBackupRestore.

        :param address: where to backup key-store or source for restoring key-store
        :param controller_id: ID of the memory to backup key-store or source memory to load key-store back
        :raises SPSDKError: If invalid address
        :raises SPSDKError: If invalid id of memory
        """
        super().__init__(self.cmd_id())
        if address < 0 or address > 0xFFFFFFFF:
            raise SPSDKError("Invalid address")
        self.header.address = address
        if controller_id < 0 or controller_id > 0xFF:
            raise SPSDKError("Invalid ID of memory")
        self.header.flags = (self.header.flags & ~self.ROM_MEM_DEVICE_ID_MASK) | (
            (controller_id << self.ROM_MEM_DEVICE_ID_SHIFT) & self.ROM_MEM_DEVICE_ID_MASK
        )
        self.header.count = (
            4  # this is useless, but it is kept for backward compatibility with elftosb
        )

    @property
    def address(self) -> int:
        """Return address where to backup key-store or source for restoring key-store."""
        return self.header.address

    @property
    def controller_id(self) -> int:
        """Return controller ID of the memory to backup key-store or source memory to load key-store back."""
        return (self.header.flags & self.ROM_MEM_DEVICE_ID_MASK) >> self.ROM_MEM_DEVICE_ID_SHIFT

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "CmdKeyStoreBackupRestore":
        """Parse command from bytes.

        :param data: Input data as bytes
        :param offset: The offset of input data
        :return: CmdKeyStoreBackupRestore object
        :raises SPSDKError: When there is invalid header tag
        """
        header = CmdHeader.parse(data, offset)
        if header.tag != cls.cmd_id():
            raise SPSDKError("Invalid header tag")
        address = header.address
        controller_id = (header.flags & cls.ROM_MEM_DEVICE_ID_MASK) >> cls.ROM_MEM_DEVICE_ID_SHIFT
        return cls(address, controller_id)  # type: ignore


class CmdKeyStoreBackup(CmdKeyStoreBackupRestore):
    """Command to backup keystore from non-volatile memory."""

    @classmethod
    def cmd_id(cls) -> EnumCmdTag:
        """Return command ID for backup operation."""
        return EnumCmdTag.WR_KEYSTORE_FROM_NV


class CmdKeyStoreRestore(CmdKeyStoreBackupRestore):
    """Command to restore keystore into non-volatile memory."""

    @classmethod
    def cmd_id(cls) -> EnumCmdTag:
        """Return command ID for restore operation."""
        return EnumCmdTag.WR_KEYSTORE_TO_NV


########################################################################################################################
# Command parser from binary format
########################################################################################################################
_CMD_CLASS: Mapping[int, Type[CmdBaseClass]] = {
    EnumCmdTag.NOP: CmdNop,
    EnumCmdTag.TAG: CmdTag,
    EnumCmdTag.LOAD: CmdLoad,
    EnumCmdTag.FILL: CmdFill,
    EnumCmdTag.JUMP: CmdJump,
    EnumCmdTag.CALL: CmdCall,
    EnumCmdTag.ERASE: CmdErase,
    EnumCmdTag.RESET: CmdReset,
    EnumCmdTag.MEM_ENABLE: CmdMemEnable,
    EnumCmdTag.PROG: CmdProg,
    EnumCmdTag.FW_VERSION_CHECK: CmdVersionCheck,
    EnumCmdTag.WR_KEYSTORE_TO_NV: CmdKeyStoreRestore,
    EnumCmdTag.WR_KEYSTORE_FROM_NV: CmdKeyStoreBackup,
}


def parse_command(data: bytes, offset: int = 0) -> CmdBaseClass:
    """Parse SB 2.x command from bytes.

    :param data: Input data as bytes
    :param offset: The offset of input data to start parsing
    :return: parsed command object
    :raises SPSDKError: Raised when there is unsupported command provided
    """
    header_tag = data[offset + 1]
    if header_tag not in _CMD_CLASS:
        raise SPSDKError(f"Unsupported command: {str(header_tag)}")
    return _CMD_CLASS[header_tag].parse(data, offset)


def get_device_id(mem_id: int) -> int:
    """Get device ID from memory ID.

    :param mem_id: memory ID
    :return: device ID
    """
    return ((mem_id) & DEVICE_ID_MASK) >> DEVICE_ID_SHIFT


def get_group_id(mem_id: int) -> int:
    """Get group ID from memory ID.

    :param mem_id: memory ID
    :return: group ID
    """
    return ((mem_id) & GROUP_ID_MASK) >> GROUP_ID_SHIFT


def get_memory_id(device_id: int, group_id: int) -> int:
    """Get memory ID from device ID and group ID.

    :param device_id: device ID
    :param group_id: group ID
    :return: memory ID
    """
    return (((group_id) << GROUP_ID_SHIFT) & GROUP_ID_MASK) | (
        ((device_id) << DEVICE_ID_SHIFT) & DEVICE_ID_MASK
    )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Secure Provisioning SDK is unified, reliable and easy to use SW library.

It's working across NXP MCU portfolio providing strong foundation from quick customer
prototyping up to production deployment.
The library allows the user to
    - connect and communicate with a device
    - configure the device; prepare
    - download and upload data
    - providing security operations.
It is delivered in a form of python library with functionality presented as CLI or GUI utilities.
"""

import logging
import logging.config
import os

from .__version__ import __version__ as version
from .exceptions import (
    SPSDKAlignmentError,
    SPSDKError,
    SPSDKIOError,
    SPSDKLengthError,
    SPSDKNotImplementedError,
    SPSDKOverlapError,
    SPSDKValueError,
)

__author__ = "NXP"
__contact__ = "michal.starecek@nxp.com"
__license__ = "BSD-3-Clause"
__version__ = version
__release__ = "alpha"

# The SPSDK behavior settings
SPSDK_DATA_FOLDER = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
SPSDK_YML_INDENT = 2

# this import has to be after SPSDK_DATA_FOLDER definition
# pylint: disable=wrong-import-position,wrong-import-order
from spsdk.utils.misc import find_file, load_configuration

try:
    logging_config_file = find_file(
        "logging.yaml", search_paths=[".", os.path.expanduser("~/.spsdk")]
    )
    config_data = load_configuration(logging_config_file)
    logging.config.dictConfig(config_data)
    print(f"Logging config loaded from {logging_config_file}")
except Exception:
    # no logging config file found
    pass



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2021 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing communication with the MCU Bootloader."""

from .commands import CommandTag, GenerateKeyBlobSelect, KeyProvUserKeyType
from .error_codes import StatusCode
from .exceptions import (
    McuBootCommandError,
    McuBootConnectionError,
    McuBootDataAbortError,
    McuBootError,
)
from .interfaces import scan_usb
from .mcuboot import McuBoot
from .memories import ExtMemId, ExtMemPropTags, MemId
from .properties import PeripheryTag, PropertyTag, Version, parse_property_value

__all__ = [
    # global methods
    "scan_usb",
    "parse_property_value",
    # classes
    "McuBoot",
    "Version",
    # enums
    "PropertyTag",
    "PeripheryTag",
    "CommandTag",
    "StatusCode",
    "ExtMemId",
    "KeyProvUserKeyType",
    # exceptions
    "McuBootError",
    "McuBootCommandError",
    "McuBootConnectionError",
]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Wrappers for various types of debuggers."""

from .debug_probe import DebugProbe
from .debug_probe_jlink import DebugProbePyLink
from .debug_probe_pyocd import DebugProbePyOCD
